import tempfile

import pytest
import torch

from .encoder import TextEmbeddingModel

TEXT_EMBEDDING_MODELS = [
    # TODO: add more models
    "intfloat/e5-small",
    "intfloat/multilingual-e5-small"
]


def mock_model_inputs(
        batch_size: int = 2,
        num_candidates: int = 4,
        seq_length: int = 10,
        vocab_size: int = 100,
    ) -> dict[str, torch.Tensor]:
    input_ids = torch.randint(0, vocab_size, (batch_size, seq_length))
    attention_mask = torch.ones((batch_size, seq_length), dtype=torch.long)
    candidates_input_ids = torch.randint(0, vocab_size, (num_candidates, seq_length))
    candidates_attention_mask = torch.ones((num_candidates, seq_length), dtype=torch.long)
    labels = torch.randint(0, num_candidates, (batch_size,))
    return {
        "input_ids": input_ids,
        "attention_mask": attention_mask,
        "candidates_input_ids": candidates_input_ids,
        "candidates_attention_mask": candidates_attention_mask,
        "labels": labels
    }


class TestTextEmbeddingModel:
    def test__init__(self) -> None:
        model = TextEmbeddingModel(TEXT_EMBEDDING_MODELS[0])
        assert isinstance(model, TextEmbeddingModel)
        assert model.distance == "inner_product"
        assert model.config == {
            "model_name_or_path": TEXT_EMBEDDING_MODELS[0],
            "pooling": "mean",
            "distance": "inner_product",
            "temperature": 1.0
        }

    @pytest.mark.parametrize("model_name", TEXT_EMBEDDING_MODELS)
    def test_encode(self, model_name: str) -> None:
        model = TextEmbeddingModel(model_name)
        model_inputs = mock_model_inputs()
        with torch.no_grad():
            mention_outputs = model.encode(
                model_inputs["input_ids"],
                model_inputs["attention_mask"],
            )
            assert mention_outputs.size() == (2, 384)

            candidate_outputs = model.encode(
                model_inputs["candidates_input_ids"],
                model_inputs["candidates_attention_mask"],
            )
            assert candidate_outputs.size() == (4, 384)

    @pytest.mark.parametrize("model_name", TEXT_EMBEDDING_MODELS)
    def test_forward(self, model_name: str) -> None:
        model = TextEmbeddingModel(model_name, pooling="mean", distance="euclidean")
        model_inputs = mock_model_inputs()
        with torch.no_grad():
            loss, logits = model(**model_inputs)
            assert loss.size() == torch.Size([])
            assert logits.size() == (2, 4)

    def test_resize_token_embeddings(self) -> None:
        model = TextEmbeddingModel(TEXT_EMBEDDING_MODELS[0], pooling="mean", distance="cosine")
        model.resize_token_embeddings(100)
        assert model.encoder.config.vocab_size == 100

    def test_save_and_load(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            model = TextEmbeddingModel(TEXT_EMBEDDING_MODELS[0], pooling="mean", distance="cosine")
            model.save_pretrained(tmpdir)
            new_model = TextEmbeddingModel.from_pretrained(tmpdir)
            assert new_model.config == model.config



import logging
import re
from argparse import ArgumentParser, Namespace

import torch

from entity_linkings import (
    ELPipeline,
    RerankerBase,
    RetrieverBase,
    get_rerankers,
    get_retrievers,
    load_dictionary,
)
from entity_linkings.utils import read_yaml

logger = logging.getLogger(__name__)

device = torch.device('cuda') if torch.cuda.is_available() else 'cpu'

def main(args: Namespace) -> None:
    dictionary = load_dictionary(args.dictionary_id_or_path, cache_dir=args.cache_dir)
    if args.retriever_config is not None:
        retriever_config = read_yaml(args.retriever_config)[args.retriever_id.lower()]
    else:
        retriever_config = {}
    if args.retriever_model_name_or_path is not None:
        retriever_config["model_name_or_path"] = args.retriever_model_name_or_path

    retriever_cls = get_retrievers(args.retriever_id)
    retriever = retriever_cls(dictionary=dictionary, config=retriever_cls.Config(**retriever_config), index_path=args.retriever_index_dir)

    model: RerankerBase|RetrieverBase = retriever
    if args.reranker_id is not None:
        if args.reranker_config is not None:
            reranker_config = read_yaml(args.reranker_config).get(args.reranker_id, {})
        else:
            reranker_config = {}
        if args.reranker_model_name_or_path is not None:
            reranker_config["model_name_or_path"] = args.reranker_model_name_or_path
        reranker_cls = get_rerankers(args.reranker_id)
        reranker = reranker_cls(retriever, config=reranker_cls.Config(**reranker_config))
        model = reranker

    pipeline = ELPipeline(model=model)

    while True:
        sentence = input('Enter: ')
        spans = []
        while True:
            match = re.search(r'\[\[(.*?)\]\]', sentence)
            if not match:
                break
            spans.append((match.start(), match.end()-4))
            mention = sentence[match.start()+2:match.end()-2]
            sentence = sentence[:match.start()] + mention + sentence[match.end():]

        predictions = pipeline.predict(sentence, spans=spans, num_candidates=args.num_candidates)

        print('================================')
        if args.reranker_id is None:
            print(f"Model Type: {args.retriever_id}")
        else:
            print(f"Model Type: {args.reranker_id} with {args.retriever_id}")
        print(f'Input Sentence: {sentence}')
        for pred in predictions:
            pred_entity = dictionary(pred.id)
            print(f'Entity Mentions: {pred.query}')
            print(f'Predicted ID: {pred.id}, Prediction: {pred_entity["name"]}')
        print('================================')


def cli_main() -> None:
    parser = ArgumentParser()
    parser.add_argument('--retriever_id', type=str, required=True, help='Name of the retriever model to use.')
    parser.add_argument('--retriever_model_name_or_path', type=str, default=None, help='Name of the retriever model to use.')
    parser.add_argument('--retriever_index_dir', type=str, default=None, help='Path to the retriever index directory.')
    parser.add_argument('--retriever_config', type=str, default=None, help='YAML-based retriever config file.')
    parser.add_argument('--reranker_id', type=str, default=None, help='Name of the model to use.')
    parser.add_argument('--reranker_model_name_or_path', type=str, default=None, help='Name of the model to use.')
    parser.add_argument('--reranker_config', type=str, default=None, help='YAML-based reranker config file.')
    parser.add_argument('--dictionary_id_or_path', type=str, default=None, help='Path to the entity dictionary file.')
    parser.add_argument('--num_candidates', type=int, default=5, help='Number of candidate entities to consider during evaluation.')
    parser.add_argument('--cache_dir', type=str, default=None, help='Path to the cache directory.')
    parser.add_argument('--top_k', type=int, default=5, help='Number of top predictions to return.')
    args = parser.parse_args()
    main(args)

if __name__ == "__main__":
    cli_main()



from dataclasses import dataclass
from typing import Optional

from datasets import Dataset
from transformers import set_seed

from entity_linkings.data_utils import EntityDictionary
from entity_linkings.trainer import EntityLinkingTrainer, TrainingArguments

from ..bm25 import BM25
from ..collator import CollatorForRetrieval
from ..dualencoder.dualencoder import compute_metrics
from ..textembedding import TEXTEMBEDDING


class E5BM25(TEXTEMBEDDING):
    """ E5BM25: Text Embedding Models for Entity Linking
    """

    @dataclass
    class Config(TEXTEMBEDDING.Config):
        model_name_or_path: Optional[str] = "intfloat/e5-base"
        candidate_pool_size: int = 40
        random_negative_sampling: bool = True
        query_type_for_candidate: str = "mention"
        language: str = "en"
        subword_tokenizer: bool = False

    def __init__(self, dictionary: EntityDictionary, config: Optional[Config] = None, index_path: Optional[str] = None) -> None:
        super().__init__(dictionary, config, index_path)

    def train(
            self,
            train_dataset: Dataset,
            eval_dataset: Optional[Dataset] = None,
            num_hard_negatives: int = 0,
            training_args: Optional[TrainingArguments] = None
        ) -> dict[str, float]:
        if num_hard_negatives > self.config.candidate_pool_size:
            raise ValueError(
                f"The number of hard negatives ({num_hard_negatives}) "
                f"cannot be larger than the candidate pool size ({self.config.candidate_pool_size})."
            )
        if training_args is None:
            training_args = TrainingArguments()
        set_seed(training_args.seed)

        bm25_config = BM25.Config(
            model_name_or_path=training_args.output_dir,
            language=self.config.language,
            n_threads=training_args.dataloader_num_workers,
            query_type_for_candidate=self.config.query_type_for_candidate,
            subword_tokenizer=self.config.model_name_or_path if self.config.subword_tokenizer else None,
        )
        bm25_retriever = BM25(self.dictionary, bm25_config)

        train_candidates = bm25_retriever.retrieve_candidates(
            train_dataset,
            only_negative=True,
            top_k=self.config.candidate_pool_size,
            batch_size=self.config.index_batch_size
        )
        train_dataset = self.preprocessor.dataset_preprocess(train_dataset, train_candidates)

        if eval_dataset is not None:
            eval_candidates = bm25_retriever.retrieve_candidates(
                eval_dataset,
                only_negative=True,
                top_k=self.config.candidate_pool_size,
                batch_size=self.config.index_batch_size
            )
            eval_dataset = self.preprocessor.dataset_preprocess(eval_dataset, eval_candidates)

        trainer = EntityLinkingTrainer(
            model=self.encoder,
            args=training_args,
            data_collator=CollatorForRetrieval(
                self.tokenizer,
                dictionary=self.dictionary,
                num_hard_negatives=num_hard_negatives,
                random_negative_sampling=self.config.random_negative_sampling,
            ),
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            compute_metrics=compute_metrics
        )
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint

        results = trainer.train(resume_from_checkpoint=checkpoint)
        trainer.log_metrics("train", results.metrics)
        if training_args.output_dir is not None:
            self.encoder.save_pretrained(training_args.output_dir)
            self.tokenizer.save_pretrained(training_args.output_dir)
            trainer.save_state()
            trainer.save_metrics("train", results.metrics)
        return results



from .base import RerankerBase
from .chatel import CHATEL
from .crossencoder import CROSSENCODER
from .extend import EXTEND
from .fevry import FEVRY
from .fusioned import FUSIONED

RERANKER_CLS: list[type[RerankerBase]] = [
    CROSSENCODER,
    CHATEL,
    FEVRY,
    EXTEND,
    FUSIONED,
]

RERANKER_ID2CLS = {
    c.__name__.lower(): c for c in RERANKER_CLS
}

__all__ = [
    "RerankerBase",
    "RERANKER_ID2CLS",
]



import abc
from dataclasses import dataclass
from typing import Any, Optional

from datasets import Dataset

from entity_linkings.candidate_retriever import RetrieverBase
from entity_linkings.trainer import TrainingArguments
from entity_linkings.utils import BaseSystemOutput


class RerankerBase(abc.ABC):
    '''
    Base class for re-ranking models
    '''
    @dataclass
    class Config:
        model_name_or_path: Optional[str] = None
        def to_dict(self) -> dict[str, Any]:
            return self.__dict__

    def __init__(self, retriever: RetrieverBase, config: Optional[Config] = None) -> None:
        self.retriever = retriever
        self.dictionary = retriever.dictionary
        self.config = config if config is not None else self.Config()

    def train(
            self,
            train_dataset: Dataset,
            eval_dataset: Optional[Dataset] = None,
            num_candidates: int = 30,
            training_args: Optional[TrainingArguments] = None
        ) -> dict[str, float]:
        raise NotImplementedError

    def evaluate(self, dataset: Dataset, num_candidates: int = 30, batch_size: int = 32, **args: int) -> dict[str, float]:
        raise NotImplementedError

    def predict(self, sentence: str, spans: list[tuple[int, int]], num_candidates: int=30) -> list[BaseSystemOutput]:
        raise NotImplementedError



import logging
import os
from dataclasses import dataclass
from typing import Optional

from transformers import AutoTokenizer

from entity_linkings.data_utils import EntityDictionary

from ..dualencoder import DUALENCODER
from .encoder import TextEmbeddingModel
from .preprocessor import TextEmbeddingPreprocessor

logger = logging.getLogger(__name__)

class TEXTEMBEDDING(DUALENCODER):
    '''
    Span-level Entity Retrieval for Text Embedding Model
    '''

    @dataclass
    class Config(DUALENCODER.Config):
        '''Retrieval configuration
        Args:
            - model (str): Pre-trained model name or path for mention and entity encoders
            - start_ent_token (str): Special token to indicate the start of entity mention
            - end_ent_token (str): Special token to indicate the end of entity mention
            - entity_token (str): Special token to indicate the entity
            - nil_token (str): Special token for NIL entity
            - max_context_length (int): Maximum sequence length for mention context
            - max_candidate_length (int): Maximum sequence length for entity context
            - pooling (str): Pooling method for obtaining fixed-size representations
            - metric (str): Similarity metric for retrieval ('inner_product', 'cosine', 'euclidean')
            - loss (str): Loss function for training ('cross_entropy' or 'triplet_margin')
            - temperature (float): Temperature for scaling similarity scores
            - prefix_context (str): Prefix to add to input texts
            - prefix_candidate (str): Prefix to add to candidate texts
            - task_description (str): Task description to add to input texts
        '''
        model_name_or_path: Optional[str] = "intfloat/e5-base"
        pooling: str = 'mean'
        prefix_context: str = "query: "
        prefix_candidate: str = "passage: "
        task_description: str = ""

    def __init__(self, dictionary: EntityDictionary, config: Optional[Config] = None, index_path: Optional[str] = None) -> None:
        self.dictionary = dictionary
        self.config = config if config is not None else self.Config()

        self.tokenizer = AutoTokenizer.from_pretrained(self.config.model_name_or_path)
        if os.path.exists(self.config.model_name_or_path):
            logger.info(f"Loading model from {self.config.model_name_or_path}")
            self.encoder = TextEmbeddingModel.from_pretrained(self.config.model_name_or_path)
        else:
            special_tokens = [self.config.ent_start_token, self.config.ent_end_token, self.config.entity_token, self.config.nil_token]
            self.tokenizer.add_special_tokens({"additional_special_tokens": special_tokens})
            self.encoder = TextEmbeddingModel(
                model_name_or_path=self.config.model_name_or_path,
                pooling=self.config.pooling,
                distance=self.config.distance,
                temperature=self.config.temperature,
            )
            self.encoder.resize_token_embeddings(len(self.tokenizer))
        self.preprocessor = TextEmbeddingPreprocessor(
            self.tokenizer, self.config.ent_start_token, self.config.ent_end_token,
            self.config.entity_token, self.config.max_context_length,
            self.config.max_candidate_length, self.config.context_window_chars
        )
        self.dictionary = self.preprocessor.dictionary_preprocess(self.dictionary)
        if index_path is not None:
            self.retriever = self.create_retriever(index_path=index_path)



import logging
from dataclasses import dataclass
from typing import Optional

from datasets import Dataset
from tqdm.auto import tqdm

from entity_linkings.data_utils import EntityDictionary
from entity_linkings.utils import BaseSystemOutput, calculate_recall_mrr

from ..base import RetrieverBase
from .indexer import MentionPriorIndexer

logger = logging.getLogger(__name__)


class PRIOR(RetrieverBase):
    '''
    Prior probability model for entity disambiguation
    '''
    @dataclass
    class Config(RetrieverBase.Config): ...

    def __init__(self, dictionary: EntityDictionary, config: Optional[Config] = None, index_path: Optional[str] = None) -> None:
        super().__init__(dictionary, config)
        if index_path is not None:
            self.retriever = self.create_retriever(index_path=index_path)

    def create_retriever(self, index_path: str) -> MentionPriorIndexer:
        retriever = MentionPriorIndexer(dictionary=self.dictionary, mention_counter_path=self.config.model_name_or_path)
        retriever.build_index(index_path=index_path)
        return retriever

    def evaluate(self, dataset: Dataset, **args: int) -> dict[str, float]:
        pbar = tqdm(total=(len(dataset)), desc='Predict')
        predictions = []
        for example in dataset:
            pbar.update()
            text = example['text']
            queries, labels = [], []
            for ent in example["entities"]:
                ent_labels = [label for label in ent['label']]
                if not ent_labels:
                    continue
                queries.append(text[ent["start"]: ent["end"]])
                labels.append(ent_labels)
            if len(queries) == 0:
                continue
            _, batch_indices = self.retriever.search_knn(queries, top_k=100)
            for i, indices in enumerate(batch_indices):
                preds = [{"id": self.dictionary(inds)["id"]} for inds in indices]
                predictions.append({"gold": labels[i], "predict": preds})
        pbar.close()
        metric = calculate_recall_mrr(predictions)
        return metric

    def predict(self, sentence: str, spans: Optional[list[tuple[int, int]]] = None, top_k: int = 5) -> list[list[BaseSystemOutput]]:
        if not spans:
            raise ValueError("Spans must be provided for PRIOR prediction.")

        queries = []
        for b, e in spans:
            queries.append(sentence[b:e])
        _, indices = self.retriever.search_knn(queries, top_k=top_k)
        all_result = []
        for i, (b, e) in enumerate(spans):
            result = []
            query = queries[i]
            for _, ind in enumerate(indices[i]):
                entry = self.dictionary(ind)
                result.append(BaseSystemOutput(query=query, start=b, end=e, id=entry['id']))
            all_result.append(result)
        return all_result

    def retrieve_candidates(self, dataset: Dataset, top_k: int = 5, only_negative: bool = False, batch_size: int = 32, **args: int) -> list[list[str]]:
        all_candidates, queries, labels = [], [], []
        pbar = tqdm(total=(len(dataset)), desc='Retrieve candidates')
        for example in dataset:
            pbar.update()
            text = example['text']
            for ent in example["entities"]:
                queries.append(text[ent["start"]: ent["end"]])
                labels.append(ent['label'])
            if len(queries) >= batch_size:
                _, batch_indices = self.retriever.search_knn(queries, top_k=top_k, ignore_ids=labels if only_negative else None)
                all_candidates.extend(batch_indices)
                queries, labels = [], []
        if len(queries) > 0:
            _, batch_indices = self.retriever.search_knn(queries, top_k=top_k, ignore_ids=labels if only_negative else None)
            all_candidates.extend(batch_indices)
        pbar.close()
        return all_candidates



from transformers import EvalPrediction


def compute_metrics(p: EvalPrediction) -> dict[str, float]:
    scores = p.predictions
    preds = scores.argmax(axis=1).ravel()
    labels = p.label_ids.ravel()
    mask = labels != -100
    preds = preds[mask]
    labels = labels[mask]

    num_corrects = (preds == labels).sum().item()
    num_golds = mask.sum().item()
    accuracy = num_corrects / num_golds if num_golds > 0 else float("nan")
    return {"accuracy": accuracy}



from importlib.resources import files

from datasets import load_dataset
from transformers import AutoTokenizer

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary

from .collator import CollatorBase

dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)


def test_CollatorBase() -> None:
    tokenizer = AutoTokenizer.from_pretrained("google-bert/bert-base-uncased")
    collator = CollatorBase(tokenizer=tokenizer)
    assert collator.tokenizer == tokenizer

    encodings = [tokenizer("This is a test."), tokenizer("Another test.")]
    batch = collator(encodings)
    assert "input_ids" in batch
    assert "attention_mask" in batch
    assert batch["input_ids"].size(0) == batch["attention_mask"].size(0) == 2
    assert batch["input_ids"].size(1) == batch["attention_mask"].size(1)



import logging
from argparse import ArgumentParser, Namespace

from entity_linkings import get_retrievers, load_dictionary
from entity_linkings.utils import read_yaml

logger = logging.getLogger(__name__)
logger.setLevel("INFO")


def build_index(args: Namespace) -> None:
    dictionary = load_dictionary(args.dictionary_id_or_path, cache_dir=args.cache_dir)
    if args.retriever_config is not None:
        retriever_config = read_yaml(args.retriever_config).get(args.retriever_id, {})
    else:
        retriever_config = {}
    if args.retriever_model_name_or_path is not None:
        retriever_config["model_name_or_path"] = args.retriever_model_name_or_path

    retriever_cls = get_retrievers(args.retriever_id)
    model = retriever_cls(dictionary=dictionary, config=retriever_cls.Config(**retriever_config))
    retriever = model.create_retriever()
    retriever.save_index(args.output_dir, ensure_ascii=False)


def cli_main() -> None:
    parser = ArgumentParser()
    parser.add_argument('--retriever_id', type=str, required=True, help='Name of the retriever model to use.')
    parser.add_argument('--retriever_model_name_or_path', type=str, default=None, help='Name of the model to use.')
    parser.add_argument('--retriever_config', type=str, default=None, help='YAML-based config file.')
    parser.add_argument('--dictionary_id_or_path', type=str, default=None, help='Path to the entity dictionary file.')
    parser.add_argument('--output_dir', type=str, default=None, help='Path to the output directory.')
    parser.add_argument("--cache_dir", type=str, default=None, help='Path to the cache directory.')
    args = parser.parse_args()
    build_index(args)

if __name__ == "__main__":
    cli_main()



import json
import os
from argparse import ArgumentParser
from typing import Any

from appendix.data_convert.utils import read_jsonl


def _convert(data: dict[str, Any]) -> dict[str, Any]:
    context = " ".join([data["left_context_text"], data["word"]])
    end = len(context)
    start = end - len(data["word"])
    context = " ".join([context, data["right_context_text"]])
    example = {
        "subset": "wikilinks_unseen",
        "id": str(data["docId"]),
        "text": context,
        "entities": [{"start": start, "end": end, "label": [str(data["wikiId"])]}],
    }
    return example


def convert_train(input_dir: str, output_dir: str) -> None:
    with open(os.path.join(output_dir, "train.jsonl"), 'w', encoding='utf-8') as f:
        for i in range(6):
            raw_examples = read_jsonl(os.path.join(input_dir, 'train', f"train_{i}.json"))
            for data in raw_examples:
                example = _convert(data)
                f.write(json.dumps(example, ensure_ascii=False) + '\n')


def convert_dev_test(input_dir: str, output_dir: str) -> list[dict[str, Any]]:
    for split in ["dev", "test"]:
        with open(os.path.join(output_dir, f"{split}.jsonl"), 'w', encoding='utf-8') as f:
            raw_examples = read_jsonl(os.path.join(input_dir, f"{split}.json"))
            examples = []
            for data in raw_examples:
                example = _convert(data)
                f.write(json.dumps(example, ensure_ascii=False) + '\n')
        return examples


def main(input_dir: str, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'wikilinks_unseen')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    convert_train(input_dir, output_dir_data)
    convert_dev_test(input_dir, output_dir_data)


if __name__ == "__main__":
    parser = ArgumentParser(description="WikiLinks Unseen Mentions Script")
    parser.add_argument("--input_dir", "-i", type=str, default='unseen_mentions', help="Path to the input text file")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    args = parser.parse_args()

    main(args.input_dir, args.output_dir)



from typing import Iterator, Optional

from datasets import Dataset
from transformers import BatchEncoding

from entity_linkings.data_utils import Preprocessor, preprocess
from entity_linkings.data_utils.utils import cut_context_window, truncate_around_mention


class DualEncoderPreprocessor(Preprocessor):
    def process_candidate(self, name: str, description: str) -> BatchEncoding:
        marked_text = name + self.entity_token + description
        encodings  = self.tokenizer(
            marked_text,
            padding=True,
            truncation=True,
            max_length=self.max_candidate_length,
        )
        return encodings

    def _process_context(self, text: str, start: int, end: int) -> str:
        context, new_start, new_end = cut_context_window(text, start, end, self.context_window_chars)
        marked_text = context[:new_start] + self.ent_start_token + context[new_start:new_end] + self.ent_end_token + context[new_end:]
        return marked_text

    def process_context(self, text: str, start: int, end: int, candidate_ids: Optional[list[str]] = None, labels: Optional[list[str]] = None) -> BatchEncoding:
        marked_text = self._process_context(text, start, end)
        input_ids = self.tokenizer.encode(marked_text, add_special_tokens=False, truncation=False)

        start_marker_idx = input_ids.index(self.start_marker_id)
        end_marker_idx = input_ids.index(self.end_marker_id)

        truncated_input_ids, _, _ = truncate_around_mention(input_ids, self.max_context_length, start_marker_idx, end_marker_idx + 1)
        final_input_ids = self.prefix_ids + truncated_input_ids + self.suffix_ids

        encodings = BatchEncoding({
            "input_ids": final_input_ids,
            "attention_mask": [1] * len(final_input_ids),
            "token_type_ids": [0] * len(final_input_ids),
        })
        if labels is not None:
            encodings["labels"] = labels
        if candidate_ids is not None:
            encodings["candidates"] = candidate_ids
        return encodings

    def dataset_preprocess(self, dataset: Dataset, candidates_ids: Optional[list[list[str]]] = None) -> Dataset:
        def _preprocess_example(examples: Dataset) -> Iterator[BatchEncoding]:
            for i, (text, entity) in enumerate(zip(examples["text"], examples["entity"])):
                candidate_ids = examples["candidates"][i] if "candidates" in examples else None
                encodings = self.process_context(text, entity['start'], entity['end'], candidate_ids, entity['label'] if 'label' in entity else None)
                yield encodings
        flatten_dataset = self.data_flatten(dataset)
        if candidates_ids is not None:
            assert len(flatten_dataset) == len(candidates_ids)
            flatten_dataset = flatten_dataset.add_column("candidates", candidates_ids)
        processed_dataset = preprocess(flatten_dataset, _preprocess_example, desc="Preprocessing")
        return processed_dataset



import json
from typing import Optional

import torch
import torch.nn as nn

from entity_linkings.utils import get_pooler, load_model


class TextEmbeddingModel(nn.Module):
    # to suppress an AttributeError when training
    _keys_to_ignore_on_save = None

    def __init__(self, model_name_or_path: str, pooling: str = "mean", distance: str = "inner_product", temperature: float = 1.0) -> None:
        """Text Embedding Model for Span Retrieval
        Args:
            model_name_or_path (str): Pre-trained model name or path
            pooling (str): Pooling method for obtaining fixed-size representations ('first', 'mean', "last")
            distance (str): Similarity metric for retrieval ('inner_product', 'cosine', 'euclidean')
            temperature (float): Temperature parameter for scaling similarity scores
        """
        super().__init__()
        self.encoder = load_model(model_name_or_path)
        self.hidden_size = self.encoder.config.hidden_size * 2 if pooling == "concat" else self.encoder.config.hidden_size
        self.pooler = get_pooler(pooling)
        if distance not in ['inner_product', 'cosine', 'euclidean']:
            raise ValueError("Distance must be 'inner_product', 'cosine', or 'euclidean'.")
        self.distance = distance
        self.temperature = temperature
        self.config = {
            "model_name_or_path": model_name_or_path,
            "pooling": pooling,
            "distance": distance,
            "temperature": temperature
        }

    def encode_candidate(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, token_type_ids: Optional[torch.Tensor] = None) -> torch.Tensor:
        return self.encode(input_ids, attention_mask, token_type_ids)

    def encode_mention(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, token_type_ids: Optional[torch.Tensor] = None) -> torch.Tensor:
        return self.encode(input_ids, attention_mask, token_type_ids)

    def encode(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, token_type_ids: Optional[torch.Tensor] = None) -> torch.Tensor:
        model_inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        if token_type_ids is not None:
            model_inputs["token_type_ids"] = token_type_ids
        outputs = self.encoder(**model_inputs).last_hidden_state
        outputs = self.pooler(outputs, attention_mask)
        return outputs

    def forward(
            self,
            input_ids: torch.Tensor,
            attention_mask: torch.Tensor,
            token_type_ids: Optional[torch.Tensor] = None,
            candidates_input_ids: Optional[torch.Tensor] = None,
            candidates_attention_mask: Optional[torch.Tensor] = None,
            candidates_token_type_ids: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None,
        ) -> tuple[Optional[torch.Tensor], torch.Tensor]:
        mention_embeddings = self.encode(input_ids, attention_mask, token_type_ids) # (bs, hidden_size)
        candidates_embeddings = self.encode(candidates_input_ids, candidates_attention_mask, candidates_token_type_ids) # (cs, hidden_size)

        if self.distance == "euclidean":
            scores = torch.cdist(mention_embeddings, candidates_embeddings, p=2)  # (bs, cs)
        else:
            if self.distance == "cosine":
                mention_embeddings = nn.functional.normalize(mention_embeddings, p=2, dim=1)
                candidates_embeddings = nn.functional.normalize(candidates_embeddings, p=2, dim=1)
            scores = mention_embeddings @ candidates_embeddings.T  # (bs, cs)

        if labels is None:
            return (None, scores)

        loss = torch.nn.functional.cross_entropy(scores, labels, reduction="mean")
        return (loss, scores)

    def resize_token_embeddings(self, new_num_tokens: int) -> None:
        self.encoder.resize_token_embeddings(new_num_tokens)

    def save_pretrained(self, save_directory: str) -> None:
        self.encoder.save_pretrained(save_directory)
        json.dump(self.config, open(f"{save_directory}/encoder_config.json", "w"), indent=2, ensure_ascii=False)

    @classmethod
    def from_pretrained(cls, load_directory: str) -> "TextEmbeddingModel":
        config = json.load(open(f"{load_directory}/encoder_config.json", "r", encoding="utf-8"))
        model = cls(**config)
        model.encoder = load_model(load_directory)
        return model



import os
from typing import Optional

import datasets

from .candidate_reranker import RERANKER_ID2CLS, RerankerBase
from .candidate_retriever import RETRIEVER_ID2CLS, RetrieverBase
from .data_utils import EntityDictionary
from .pipeline import ELPipeline


def get_retriever_ids() -> list[str]:
    '''Generate a list of ids with the class name in lower case.
    '''
    ids = list(RETRIEVER_ID2CLS.keys())
    return ids


def get_reranker_ids() -> list[str]:
    '''Generate a list of ids with the class name in lower case.
    '''
    ids = list(RERANKER_ID2CLS.keys())
    return ids


def get_model_ids() -> list[str]:
    '''Generate a list of ids with the class name in lower case.
    '''
    ids = list(RETRIEVER_ID2CLS.keys()) + list(RERANKER_ID2CLS.keys())
    return ids


def load_dictionary(
        dictionary_name_or_path: str,
        nil_id: str = "-1",
        nil_name: str = "[NIL]",
        nil_description: str = "[NIL] is an entity that does not exist in this dictionary.",
        default_description: str = """{name} is an entity in this dictionary.""",
        cache_dir: Optional[str|os.PathLike] = None,
    ) -> EntityDictionary:
    '''Generate a dictionary of ids and classes with the class name in lower case as the key.
    '''
    if os.path.isfile(dictionary_name_or_path):
        dictionary = datasets.load_dataset("json", data_files=dictionary_name_or_path, cache_dir=cache_dir, split="train")
    else:
        if not dictionary_name_or_path.startswith("naist-nlp/"):
            dictionary_name_or_path = f"naist-nlp/{dictionary_name_or_path}"
        dictionary = datasets.load_dataset(dictionary_name_or_path, 'dictionary', split='kb', cache_dir=cache_dir)

    return EntityDictionary(
        dictionary=dictionary,
        nil_id=nil_id,
        nil_name=nil_name,
        nil_description=nil_description,
        default_description=default_description,
        cache_dir=cache_dir,
    )


def get_retrievers(name: str) -> type[RetrieverBase]:
    '''Generate a retriever model class.
    If without_span is True, use SentenceRetrieval class.
    Otherwise, use Retrieval class.
    '''
    if name not in get_retriever_ids():
        raise ValueError(f"The id should be one of {get_retriever_ids()}.")
    return RETRIEVER_ID2CLS[name]


def get_rerankers(name: str) -> type[RerankerBase]:
    '''Generate a dictionary of ids and classes with the class name in lower case as the key.
    '''
    if name not in get_reranker_ids():
        raise ValueError(f"The id should be one of {get_reranker_ids()}.")
    return RERANKER_ID2CLS[name]


__all__ = [
    "get_retriever_ids",
    "get_reranker_ids",
    "get_model_ids",
    "load_dictionary",
    "get_retrievers",
    "get_rerankers",
    "EntityDictionary",
    "RetrieverBase",
    "RerankerBase",
    "ELPipeline",
]



import tempfile
from importlib.resources import files

import numpy as np
import pytest
from datasets import load_dataset
from faiss import (
    METRIC_INNER_PRODUCT,
    METRIC_L2,
    IndexFlatIP,
    IndexFlatL2,
    IndexHNSWFlat,
)
from transformers import AutoTokenizer

import assets as test_data
from entity_linkings import load_dictionary

from .encoder import DualBERTModel
from .indexer import DenseRetriever
from .preprocessor import DualEncoderPreprocessor

model_name = 'google-bert/bert-base-uncased'
model = DualBERTModel(model_name_or_path=model_name)
tokenizer = AutoTokenizer.from_pretrained(model_name)
special_tokens = {"additional_special_tokens": ["[ENT]", "[START_ENT]", "[END_ENT]"]}
tokenizer.add_special_tokens(special_tokens)
model.resize_token_embeddings(len(tokenizer))

test_dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": test_dataset_path})['test']
preprocessor = DualEncoderPreprocessor(tokenizer, "[START_ENT]", "[END_ENT]", "[ENT]", 128, 128, 500)
processed_dictionary = preprocessor.dictionary_preprocess(dictionary)


class TestDenseRetriever:
    def test___init__(self) -> None:
        retriever = DenseRetriever(
            model = model,
            tokenizer=tokenizer,
            dictionary=processed_dictionary,
        )
        assert isinstance(retriever, DenseRetriever)
        assert retriever.model is model
        assert retriever.dictionary is processed_dictionary
        assert retriever.vector_size == model.hidden_size

    @pytest.mark.parametrize("metric",["inner_product", "cosine", "euclidean"])
    @pytest.mark.parametrize("use_hnsw",[True, False])
    def test_initialize(self, use_hnsw: bool, metric: str) -> None:
        retriever = DenseRetriever(
            model=model,
            tokenizer=tokenizer,
            dictionary=processed_dictionary,
            use_hnsw=use_hnsw,
            metric=metric,
        )
        retriever._initialize()
        if use_hnsw:
            if metric == "euclidean":
                assert isinstance(retriever.index, IndexHNSWFlat) and retriever.index.metric_type == METRIC_L2
            else:
                assert isinstance(retriever.index, IndexHNSWFlat) and retriever.index.metric_type == METRIC_INNER_PRODUCT
        else:
            if metric == "euclidean":
                assert isinstance(retriever.index, IndexFlatL2)
            else:
                assert isinstance(retriever.index, IndexFlatIP)
        assert hasattr(retriever, "meta_ids_to_keys") and isinstance(retriever.meta_ids_to_keys, dict)
        assert len(retriever.meta_ids_to_keys) == retriever.index.ntotal == 0

    def test_build_index(self) -> None:
        retriever = DenseRetriever(
            model=model,
            tokenizer=tokenizer,
            dictionary=processed_dictionary,
        )
        retriever.build_index()
        assert isinstance(retriever.index, IndexFlatIP)
        assert isinstance(retriever.meta_ids_to_keys, dict)
        assert len(retriever.meta_ids_to_keys) == retriever.index.ntotal == len(dictionary)

    @pytest.mark.parametrize("top_k", [0, 2, 5, 10])
    def test_search_knn(self, top_k: int) -> None:
        retriever = DenseRetriever(
            model=model,
            tokenizer=tokenizer,
            dictionary=processed_dictionary,
            batch_size=2,
        )
        retriever.build_index()
        for data in dataset:
            query = data['text']
            if top_k <= 0:
                with pytest.raises(RuntimeError) as re:
                    retriever.search_knn(query, top_k)
                assert isinstance(re.value, RuntimeError)
                assert str(re.value) == "K is zero or under zero."
                break
            else:
                distances, indices = retriever.search_knn(query, top_k=top_k)
                assert isinstance(distances, np.ndarray) and isinstance(indices, list)
                if top_k > len(dictionary):
                    assert distances.shape[0] == len(indices) == 1
                    assert distances.shape[1] == len(indices[0]) == len(dictionary)
                else:
                    assert distances.shape[0] == len(indices) == 1
                    assert distances.shape[1] == len(indices[0]) == top_k

    @pytest.mark.parametrize("top_k", [2, 4])
    def test_search_knn_negatives(self, top_k: int) -> None:
        retriever = DenseRetriever(
            model=model,
            tokenizer=tokenizer,
            dictionary=processed_dictionary,
        )
        retriever.build_index()
        for text, entities in zip(dataset['text'], dataset['entities']):
            for ent in entities:
                if 'label' not in ent:
                    continue
                query = text
                labels = ent["label"]
                _, indices = retriever.search_knn(query, top_k, ignore_ids=labels)
                for i, inds in enumerate(indices):
                    assert len(inds) == top_k
                    for ind in inds:
                        assert ind not in labels[i]

    def test_save_and_load(self) -> None:
        retriever = DenseRetriever(
            model=model,
            tokenizer=tokenizer,
            dictionary=processed_dictionary,
        )
        with tempfile.TemporaryDirectory() as tmpdir:
            retriever.build_index(tmpdir)
            loaded_retriever = DenseRetriever(
                model=model,
                tokenizer=tokenizer,
                dictionary=processed_dictionary,
            )
            loaded_retriever.build_index(tmpdir)
            assert retriever.dictionary and loaded_retriever.dictionary
            assert retriever.meta_ids_to_keys == loaded_retriever.meta_ids_to_keys
            assert retriever.dictionary.id_to_index == loaded_retriever.dictionary.id_to_index

    def test_len(self) -> None:
        retriever = DenseRetriever(
            model=model,
            tokenizer=tokenizer,
            dictionary=processed_dictionary,
        )
        retriever.build_index()
        assert len(retriever) == len(dictionary)



import json
import os
from argparse import ArgumentParser
from itertools import accumulate
from typing import Any, Iterable, Union

from appendix.data_convert.utils import _conll_to_example


def read_conll(
        file: Union[str, bytes, os.PathLike],
        delimiter: str = '\t',
    ) -> Iterable[tuple[str, list[dict[str, Any]]]]:
    sentences: list[dict[str, Any]] = []
    words: list[str] = []
    labels: list[str] = []
    links: list[str] = []
    id = ""

    with open(file, mode="r", encoding="utf-8") as f:
        for line in f:
            line = line.rstrip()
            if line.startswith("-DOCSTART-"):
                if sentences:
                    yield id, sentences
                    sentences = []
                id = line[11:].strip()[1:-1]
            elif not line:
                if words:
                    sentences.append(_conll_to_example(words, labels, links))
                    words = []
                    labels = []
                    links = []
            else:
                cols = line.split(delimiter)
                words.append(cols[0])
                if len(cols) == 1:
                    labels.append('O')
                    links.append('')
                else:
                    labels.append(cols[1])
                    links.append(cols[3] if len(cols) == 4 else cols[5])
    if sentences:
        yield id, sentences


def read_tsv_file(file_path: str) -> dict[str, list[dict]]:
    train, testa, testb = [], [], []
    for id, data in read_conll(file_path, delimiter='\t'):
        texts = [d['text'] for d in data]
        context = " ".join(texts)
        cumsum_lengths = list(accumulate([0] + [len(t) + 1 for t in texts]))
        entities = []
        for i, d in enumerate(data):
            for ent in d['entities']:
                start = ent['start'] + cumsum_lengths[i]
                end = ent['end'] + cumsum_lengths[i]
                title = ["-1"] if ent['title'] == ['--NME--'] else ent['title']
                assert context[start:end] == ent["text"]
                entities.append({"start": start, "end": end, "label": title})
        example = {
            "id": id,
            "text": context,
            "entities": entities,
        }
        if 'testa' in id:
            testa.append(example)
        elif 'testb' in id:
            testb.append(example)
        else:
            train.append(example)
    print(f"Train size: {len(train)}")
    print(f"TestA size: {len(testa)}")
    print(f"TestB size: {len(testb)}")
    return {"train": train, "testa": testa, "testb": testb}

def main(input_file: str, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'aida')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    data = read_tsv_file(input_file)
    for split in ['train', 'testa', 'testb']:
        with open(os.path.join(output_dir_data, f"{split}.jsonl"), 'w', encoding='utf-8') as f:
            for example in data[split]:
                f.write(json.dumps(example, ensure_ascii=False) + "\n")


if __name__ == "__main__":
    parser = ArgumentParser(description="KILT Entity Disambiguation Script")
    parser.add_argument("--input_file", "-i", type=str, default="AIDA-YAGO2-dataset.tsv", help="Path to the input text file")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    args = parser.parse_args()

    main(args.input_file, args.output_dir)



import tempfile

import pytest
import torch
from transformers import (
    BertModel,
    DebertaV2Model,
    ModernBertModel,
    RobertaModel,
    XLMRobertaModel,
)

from .encoder import Encoder

MODELS = [
    "google-bert/bert-base-uncased",
    "FacebookAI/xlm-roberta-base",
    "microsoft/deberta-v3-base",
    "FacebookAI/roberta-base",
    "answerdotai/ModernBERT-base",
]

def mock_input(
        batch_size: int = 2,
        num_candidates: int = 3,
        seq_length: int = 10,
        vocab_size: int = 100
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    import torch
    input_ids = torch.randint(0, vocab_size, (batch_size, num_candidates, seq_length))
    attention_mask = torch.ones((batch_size, num_candidates, seq_length), dtype=torch.long)
    labels = torch.randint(0, num_candidates, (batch_size, ), dtype=torch.long)
    return input_ids, attention_mask, labels

class TestEncoder:
    @pytest.mark.parametrize("model_name", MODELS)
    def test__init__(self, model_name: str) -> None:
        model = Encoder(model_name)
        assert isinstance(model, Encoder)
        assert hasattr(model, "encoder") and hasattr(model, "projection") and hasattr(model, "pooler")
        assert isinstance(model.encoder, (BertModel, RobertaModel, DebertaV2Model, XLMRobertaModel, ModernBertModel))

    def test_resize_token_embeddings(self) -> None:
        model = Encoder(MODELS[0])
        old_embeddings = model.encoder.get_input_embeddings()
        old_num_tokens, embedding_dim = old_embeddings.weight.size()
        new_num_tokens = old_num_tokens + 10
        model.resize_token_embeddings(new_num_tokens)
        new_embeddings = model.encoder.get_input_embeddings()
        assert new_embeddings.weight.size() == (new_num_tokens, embedding_dim)

    @pytest.mark.parametrize("model_name", MODELS)
    def test_encode(self, model_name: str) -> None:
        model = Encoder(model_name)
        input_ids, attention_mask, _ = mock_input()
        bs, cs, length = input_ids.size()
        input_ids = input_ids.view(bs * cs, length)
        attention_mask = attention_mask.view(bs * cs, length)

        with torch.no_grad():
            pooled_output = model.encode(input_ids=input_ids, attention_mask=attention_mask)
        assert pooled_output.size(0) == input_ids.size(0)
        assert pooled_output.size(1) == model.encoder.config.hidden_size

    @pytest.mark.parametrize("model_name", MODELS)
    def test_forward(self, model_name: str) -> None:
        model = Encoder(model_name)
        input_ids, attention_mask, labels = mock_input()
        with torch.no_grad():
            loss, scores = model(input_ids=input_ids, attention_mask=attention_mask, labels=labels)
        assert isinstance(loss, torch.Tensor)
        assert scores.size(0) == input_ids.size(0)
        assert scores.size(1) == input_ids.size(1)

    @pytest.mark.parametrize("model_name", [MODELS[0]])
    def test_save_and_load(self, model_name: str) -> None:
        model = Encoder(model_name)
        with tempfile.TemporaryDirectory() as tmpdir:
            model.save_pretrained(tmpdir)
            loaded_model = Encoder.from_pretrained(tmpdir)
            assert isinstance(loaded_model, Encoder)



import tempfile
from importlib.resources import files

import pytest
from datasets import load_dataset
from transformers import BertTokenizerFast
from transformers.trainer_utils import TrainOutput

import assets as test_data
from entity_linkings import load_dictionary
from entity_linkings.trainer import TrainingArguments
from entity_linkings.utils import BaseSystemOutput

from .dualencoder import DUALENCODER
from .encoder import DualBERTModel

BERT_MODELS = ["google-bert/bert-base-uncased"]

dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']


@pytest.mark.retriever_dual_encoder
class TestSpanEntityRetrieverForDualEncoder:
    def test__init__(self) -> None:
        model = DUALENCODER(
            dictionary=dictionary,
            config=DUALENCODER.Config(model_name_or_path=BERT_MODELS[0])
        )
        assert isinstance(model, DUALENCODER)
        assert model.config.ent_start_token in model.tokenizer.all_special_tokens
        assert model.config.ent_start_token in model.tokenizer.all_special_tokens
        assert model.config.entity_token in model.tokenizer.all_special_tokens
        assert model.config.nil_token in model.tokenizer.all_special_tokens
        assert isinstance(model.encoder, DualBERTModel)
        assert isinstance(model.tokenizer, BertTokenizerFast)
        assert model.config.model_name_or_path == BERT_MODELS[0]
        for processed in model.dictionary:
            assert "encoding" in processed
            assert "input_ids" in processed["encoding"]
            assert "attention_mask" in processed["encoding"]

    def test_train(self) -> None:
        model = DUALENCODER(dictionary=dictionary)
        with tempfile.TemporaryDirectory() as tmpdir:
            result = model.train(
                train_dataset=dataset,
                eval_dataset=dataset,
                num_hard_negatives=2,
                training_args=TrainingArguments(
                    output_dir=tmpdir,
                    num_train_epochs=1,
                    per_device_train_batch_size=2,
                    per_device_eval_batch_size=2,
                    logging_strategy="no",
                    save_strategy="no",
                    eval_strategy="no",
                    remove_unused_columns=False,
                    eval_on_start=True
                )
            )
            assert isinstance(result, TrainOutput)
            assert hasattr(result, 'metrics')

    def test_evaluate(self) -> None:
        model = DUALENCODER(dictionary=dictionary)
        metrics = model.evaluate(dataset)
        assert 'recall@1' in metrics
        assert 'recall@10' in metrics
        assert 'recall@50' in metrics
        assert 'recall@100' in metrics
        assert 'mrr@100' in metrics

    def test_predict(self) -> None:
        model = DUALENCODER(dictionary=dictionary)
        sentence = "Steve Jobs was found Apple."
        spans = [(21, 26)]
        top_k = 3
        predictions = model.predict(sentence, spans, top_k)
        assert isinstance(predictions, list)
        assert len(predictions) == len(spans)
        for preds in predictions:
            assert isinstance(preds, list)
            assert len(preds) == min(top_k, len(dictionary))
            for pred in preds:
                assert isinstance(pred, BaseSystemOutput)
                assert pred.id is not None and pred.start is not None and pred.end is not None

    def test_retrieve_candidates(self) -> None:
        model = DUALENCODER(dictionary=dictionary)
        top_k = 3
        candidate_lists = model.retrieve_candidates(dataset, top_k=top_k, batch_size=1, negative=True)
        assert isinstance(candidate_lists, list)
        assert len(candidate_lists) == len(dataset)
        for candidates in candidate_lists:
            assert isinstance(candidates, list)
            assert len(candidates) == top_k
            for candidate in candidates:
                assert isinstance(candidate, str)



import json
import os
from argparse import ArgumentParser
from pathlib import Path

import requests
from datasets import load_dataset
from tqdm.auto import tqdm

TASKS = ['wned', 'cweb', 'aidayago2']
base_url = "http://dl.fbaipublicfiles.com/KILT"


def _convert(data: dict, subset: str) -> dict:
    mention = data["meta"]["mention"].replace(u'\xa0', ' ')
    left_context = data["meta"]["left_context"].replace(u'\xa0', ' ')
    right_context = data["meta"]["right_context"].replace(u'\xa0', ' ')

    text = f"{left_context} {mention} {right_context}".strip()
    start = len(left_context) + 1 if left_context else 0
    end = start + len(mention)

    assert text[start: end] == mention, (text[start: end], mention)
    if data['output'] != []:
        assert len(data['output']) == 1 and len(data['output'][0]['provenance']) == 1
        wikipedia_id = str(data['output'][0]['provenance'][0]['wikipedia_id'])
    else:
        wikipedia_id = "-1"

    example = {
        "subset": subset,
        "id": data["id"],
        "text": text,
        "entities": [{
            "start": start,
            "end": end,
            "label": [wikipedia_id],
        }],
    }
    return example


def convert_dictionary(output_dir: str) -> None:
    url = os.path.join(base_url, "kilt_knowledgesource.json")
    urlData = requests.get(url, stream=True)
    pbar = tqdm(total=5903530, desc="Downloading dictionary")
    with open(os.path.join(output_dir, "dictionary.jsonl"), 'w', encoding='utf-8') as f:
        for chunk in urlData.iter_lines():
            pbar.update()
            line = json.loads(chunk)
            description = "" if len(line['text']) < 2 else line['text'][1].strip()
            entity = {"id": str(line['wikipedia_id']), "title": line['wikipedia_title'], "text": description}
            f.write(json.dumps(entity, ensure_ascii=False) + "\n")
    pbar.close()
    print(f"Converted dictionary to {output_dir}/dictionary.jsonl")


def convert_to_additional_data(output_dir: str) -> None:
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    for split in ["train", "dev"]:
        pbar = tqdm(total=9000000 if split == 'train' else 10000, desc=f"Downloading {split} additional data")
        url = os.path.join(base_url, f"blink-{split}-kilt.jsonl")
        urlData = requests.get(url, stream=True)
        with open(os.path.join(output_dir, f"{split}.jsonl"), 'w', encoding='utf-8') as f:
            for chunk in urlData.iter_lines():
                pbar.update()
                line = json.loads(chunk)
                example = _convert(line, subset="wikipedia")
                f.write(json.dumps(example, ensure_ascii=False) + "\n")
        pbar.close()
    print(f"Converted additional data to {output_dir}")


def convert_dataset(task: str, output_dir: str) -> None:
    kilt_task = load_dataset("facebook/kilt_tasks", task)
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    for split in kilt_task.keys():
        pbar = tqdm(total=len(kilt_task[split]), desc=f"Converting {task} {split}")
        with open(os.path.join(output_dir, f"{split}.jsonl"), 'w', encoding='utf-8') as f:
            for chunk in kilt_task[split]:
                pbar.update()
                example = _convert(chunk, subset=task)
                f.write(json.dumps(example, ensure_ascii=False) + "\n")
        pbar.close()
    print(f"Converted {task} to {output_dir}")


def main(output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'kilt')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    convert_dictionary(output_dir_data)
    convert_to_additional_data(output_dir_data)
    for task in TASKS:
        convert_dataset(task, os.path.join(output_dir_data, task))


if __name__ == "__main__":
    parser = ArgumentParser(description="KILT Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    args = parser.parse_args()

    main(args.output_dir)



import json
import os
from argparse import ArgumentParser
from typing import Any

from appendix.data_convert.utils import read_jsonl

DOMAIN_SPLITS = {
    "train": [
        "american_football",
        "doctor_who",
        "fallout",
        "final_fantasy",
        "military",
        "pro_wrestling",
        "starwars",
        "world_of_warcraft",
    ],
    "dev": [
        "coronation_street",
        "muppets",
        "ice_hockey",
        "elder_scrolls",
    ],
    "test": [
        "forgotten_realms",
        "lego",
        "star_trek",
        "yugioh"
    ]
}

def get_documents(input_dir: str, split: str) -> dict[str, Any]:
    documents = {}
    for domain in DOMAIN_SPLITS[split]:
        file_path = os.path.join(input_dir, f"documents/{domain}.json")
        for line in read_jsonl(file_path):
            documents[line["document_id"]] = {"title": line["title"], "text": line["text"], "domain": domain}
    if split == "train":
        assert len(documents) == 332632, f"Expected 332632 documents for train, got {len(documents)}"
    elif split == "dev":
        assert len(documents) == 89549, f"Expected 89549 documents for dev, got {len(documents)}"
    elif split == "test":
        assert len(documents) == 70140, f"Expected 70140 documents for test, got {len(documents)}"
    return documents


def convert_dictionary(raw_datasets: dict[str, Any], output_dir: str) -> None:
    with open(os.path.join(output_dir, "dictionary.jsonl"), 'w', encoding='utf-8') as f:
        for split in raw_datasets:
            for doc_id, doc in raw_datasets[split].items():
                f.write(json.dumps({"id": doc_id, "title": doc["title"], "text": doc["text"], "domain": doc["domain"]}, ensure_ascii=False) + "\n")


def get_mention_dictionary(input_dir: str, split: str) -> dict[str, Any]:
    file_path = os.path.join(input_dir, f"mentions/{split}.json")
    mention_dict: dict[str, Any] = {}
    for line in read_jsonl(file_path):
        context_document_id = line["context_document_id"]
        if context_document_id not in mention_dict:
            mention_dict[context_document_id] = []
        mention_dict[context_document_id].append(line)
    return mention_dict


def convert_dataset(dataset: dict[str, Any], mention_dict: dict[str, Any], output_path: str) -> None:
    mention_count = 0
    with open(output_path, 'w', encoding='utf-8') as f:
        for doc_id, doc in dataset.items():
            if doc_id not in mention_dict:
                continue
            example = {"subset": doc["domain"], "id": doc_id, "text": doc["text"], "entities": []}
            mentions = mention_dict[doc_id]
            mention_count += len(mentions)
            for mention in mentions:
                start, end, label = mention["start_index"], mention["end_index"], mention["label_document_id"]
                assert ' '.join(doc["text"].split()[start:end+1]) == mention["text"]

                char_start = len(' '.join(doc["text"].split()[:start])) + 1 if start > 0 else 0
                char_end = len(' '.join(doc["text"].split()[:end+1]))
                assert doc["text"][char_start:char_end] == mention["text"]

                example["entities"].append({
                    "start": char_start,
                    "end": char_end,
                    "text": mention["text"],
                    "label": [label],
                    "category": mention["category"]
                })
            f.write(json.dumps(example, ensure_ascii=False) + "\n")
    assert mention_count == sum(len(v) for v in mention_dict.values())


def main(input_dir: str, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'zeshel')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    raw_datasets = {split: get_documents(input_dir, split) for split in ["train", "dev", "test"]}
    convert_dictionary(raw_datasets, output_dir_data)
    for split in ["train", "heldout_train_seen", "heldout_train_unseen"]:
        mention_dict = get_mention_dictionary(input_dir, split)
        if split == "train":
            assert sum(len(v) for v in mention_dict.values()) == 49275
        else:
            assert sum(len(v) for v in mention_dict.values()) == 5000
        convert_dataset(raw_datasets["train"], mention_dict, os.path.join(output_dir_data, f"{split}.jsonl"))

    for split in ["dev", "test"]:
        mention_dict = get_mention_dictionary(input_dir, split if split == "test" else "val")
        assert sum(len(v) for v in mention_dict.values()) == 10000
        convert_dataset(raw_datasets[split], mention_dict, os.path.join(output_dir_data, f"{split}.jsonl"))


if __name__ == "__main__":
    parser = ArgumentParser(description="ZeshEL Script")
    parser.add_argument("--input_dir", "-i", type=str, default='zeshel', help="Path to the input text file")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    args = parser.parse_args()

    main(args.input_dir, args.output_dir)




import pytest
import torch
from transformers import AutoModelForSeq2SeqLM

from .reader import CheckpointWrapper, EncoderWrapper, FusionDecoder

MODELS = ["google/flan-t5-small"]


def mock_input(
        batch_size: int = 2,
        num_candidates: int = 3,
        seq_length: int = 10,
        vocab_size: int = 100
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    input_ids = torch.randint(0, vocab_size, (batch_size, num_candidates, seq_length))
    attention_mask = torch.ones((batch_size, num_candidates, seq_length), dtype=torch.long)
    labels = torch.randint(0, vocab_size, (batch_size, seq_length))
    return input_ids, attention_mask, labels


class TestEncoderWrapper:
    @pytest.mark.parametrize("model_name", MODELS)
    @pytest.mark.parametrize("use_checkpoint", [True, False])
    def test__init__(self, model_name: str, use_checkpoint: bool) -> None:
        t5 = AutoModelForSeq2SeqLM.from_pretrained(model_name)
        encoder_wrapper = EncoderWrapper(t5.get_encoder(), use_checkpoint=use_checkpoint)
        assert isinstance(encoder_wrapper, EncoderWrapper)
        assert encoder_wrapper.main_input_name == "input_ids"
        assert hasattr(encoder_wrapper, "encoder")
        for module in encoder_wrapper.encoder.block:
            assert isinstance(module, CheckpointWrapper)
            assert hasattr(module, "forward")
            assert module.use_checkpoint == use_checkpoint

    @pytest.mark.parametrize("model_name", MODELS)
    @pytest.mark.parametrize("use_checkpoint", [True, False])
    def test_forward(self, model_name: str, use_checkpoint: bool) -> None:
        t5 = AutoModelForSeq2SeqLM.from_pretrained(model_name)
        encoder_wrapper = EncoderWrapper(t5.get_encoder(), use_checkpoint=use_checkpoint)
        input_ids, attention_mask, _ = mock_input()
        encoder_wrapper.n_passages = input_ids.size(1)
        input_ids = input_ids.view(input_ids.size(0), -1)
        attention_mask = attention_mask.view(attention_mask.size(0), -1)
        outputs = encoder_wrapper(input_ids=input_ids, attention_mask=attention_mask)
        assert hasattr(outputs, "last_hidden_state")


class TestFusionDecoder:
    @pytest.mark.parametrize("model_name", MODELS)
    def test__init__(self, model_name: str) -> None:
        t5 = AutoModelForSeq2SeqLM.from_pretrained(model_name)
        model = FusionDecoder(t5.config)
        model.load_t5(t5.state_dict())
        assert isinstance(model, FusionDecoder)
        assert isinstance(model.encoder, EncoderWrapper)

    @pytest.mark.parametrize("model_name", MODELS)
    def test_forward(self, model_name: str) -> None:
        t5 = AutoModelForSeq2SeqLM.from_pretrained(model_name)
        model = FusionDecoder(t5.config)
        model.load_t5(t5.state_dict())

        input_ids, attention_mask, labels = mock_input()
        outputs = model(input_ids=input_ids, attention_mask=attention_mask, labels=labels)
        assert hasattr(outputs, "loss")
        assert hasattr(outputs, "logits")

    @pytest.mark.parametrize("model_name", MODELS)
    def test_generate(self, model_name: str) -> None:
        t5 = AutoModelForSeq2SeqLM.from_pretrained(model_name)
        model = FusionDecoder(t5.config)
        model.load_t5(t5.state_dict())
        input_ids, attention_mask, _ = mock_input()
        generated_ids = model.generate(input_ids=input_ids, attention_mask=attention_mask)
        assert generated_ids.size(0) == input_ids.size(0)
        assert 3 <= generated_ids.size(1) <= 201

    @pytest.mark.parametrize("model_name", MODELS)
    def test_save_pretrained(self, model_name: str, tmp_path: pytest.TempPathFactory) -> None:
        t5 = AutoModelForSeq2SeqLM.from_pretrained(model_name)
        model = FusionDecoder(t5.config)
        model.load_t5(t5.state_dict())

        save_dir = tmp_path / "fusioned_model"
        model.save_pretrained(save_dir)
        loaded_model = FusionDecoder.from_pretrained(save_dir)
        assert isinstance(loaded_model, FusionDecoder)
        assert isinstance(loaded_model.encoder, EncoderWrapper)
        for block in loaded_model.encoder.encoder.block:
            assert isinstance(block, CheckpointWrapper)



import tempfile
from importlib.resources import files

import pytest
from datasets import load_dataset
from transformers.trainer_utils import TrainOutput

import assets as test_data
from entity_linkings import load_dictionary
from entity_linkings.trainer import TrainingArguments

from ..textembedding.encoder import TextEmbeddingModel
from .e5bm25 import E5BM25

MODELS = ["intfloat/e5-base"]

dataset_path = str(files(test_data).joinpath("dataset_toy_wo_candidates.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']


@pytest.mark.retriever_e5bm25
class TestE5BM25:
    @pytest.mark.parametrize("model_name", MODELS)
    def test__init__(self, model_name: str) -> None:
        model = E5BM25(
            dictionary=dictionary,
            config=E5BM25.Config(model_name_or_path=model_name)
        )
        assert isinstance(model, E5BM25)
        assert isinstance(model.encoder, TextEmbeddingModel)
        assert model.config.candidate_pool_size == 40
        assert model.config.random_negative_sampling is True
        assert model.config.model_name_or_path == model_name

    @pytest.mark.parametrize("model_name", MODELS)
    def test_train(self, model_name: str) -> None:
        model = E5BM25(
            dictionary=dictionary,
            config=E5BM25.Config(
                candidate_pool_size=3,
                model_name_or_path=model_name
            )
        )
        with tempfile.TemporaryDirectory() as tmpdir:
            result = model.train(
                    train_dataset=dataset,
                    eval_dataset=dataset,
                    num_hard_negatives=1,
                    training_args=TrainingArguments(
                    output_dir=tmpdir,
                    num_train_epochs=1,
                    per_device_train_batch_size=2,
                    per_device_eval_batch_size=2,
                    logging_strategy="no",
                    save_strategy="no",
                    eval_strategy="no",
                    remove_unused_columns=False,
                    eval_on_start=True
                )
            )
            assert isinstance(result, TrainOutput)
            assert hasattr(result, 'metrics')



import json
import os
from typing import Any, Iterable, Iterator, Union

import duckdb
import requests


def get_wikipedia_summary(title: str, lang: str = 'en') -> dict[str, str]:
    ENDPOINT = f'https://{lang}.wikipedia.org/api/rest_v1/page/summary/'
    HEADERS = {'User-Agent': 'EntityLinkingBenchmark'}

    response = requests.get(ENDPOINT + title, headers=HEADERS)
    if response.status_code == 200:
        data = response.json()
        return {'pageid': str(data.get('pageid', -1)), 'wikibase_id': str(data.get('wikibase_item', -1))}
    else:
        return {'pageid': str(-1), 'wikibase_id': str(-1)}


class WikiMapper:
    def __init__(self, db_path: str) -> None:
        self.con = duckdb.connect(db_path, read_only=True)

    def find_by_title(self, title: str, return_redirect: bool = False) -> dict[str, str] | None:
        # Replace space with underscore to match Wikipedia's internal format
        internal_title = title.replace(" ", "_")

        query = "SELECT page_id, title, wikidata_id, is_redirect, redirect_to FROM wiki_map WHERE title = ?"
        result = self.con.execute(query, [internal_title]).fetchone()

        if result:
            if return_redirect:
                return {
                    "page_id": result[0],
                    "title": result[1],
                    "wikidata_id": result[2],
                    "is_redirect": result[3],
                    "redirect_to": result[4]
                }
            else:
                query_page = "SELECT page_id, title, wikidata_id, is_redirect, redirect_to FROM wiki_map WHERE page_id = ?"
                while True:
                    if not result[3]:
                        break
                    result = self.con.execute(query_page, [result[4]]).fetchone()
                    assert result is not None
                return {"page_id": result[0], "wikidata_id": result[2]}
        return None

    def find_by_page_id(self, page_id: str, return_redirect: bool = False) -> dict[str, str] | None:
        query = "SELECT page_id, title, wikidata_id, is_redirect, redirect_to FROM wiki_map WHERE page_id = ?"
        result = self.con.execute(query, [page_id]).fetchone()

        if result:
            if return_redirect:
                return {
                    "page_id": result[0],
                    "title": result[1],
                    "wikidata_id": result[2],
                    "is_redirect": result[3],
                    "redirect_to": result[4]
                }
            else:
                while True:
                    if not result[3]:
                        break
                    result = self.con.execute(query, [result[4]]).fetchone()
                    assert result is not None
                return {"page_id": result[0], "wikidata_id": result[2]}
        return None

    def find_by_wikidata_id(self, wikidata_id: str) -> list[dict[str, str]]:
        query = "SELECT page_id, title, wikidata_id, is_redirect, redirect_to FROM wiki_map WHERE wikidata_id = ?"
        results = self.con.execute(query, [wikidata_id]).fetchall()

        return [
            {
                "page_id": result[0],
                "title": result[1],
                "wikidata_id": result[2],
                "is_redirect": result[3],
                "redirect_to": result[4],
            } for result in results
        ]

    def close(self) -> None:
        self.con.close()


def read_conll(
        file: Union[str, bytes, os.PathLike],
        delimiter: str = ' ',
        word_column: int = 0,
        tag_column: int = 1,
        link_column: int = 2
    ) -> Iterable[tuple[str, list[dict[str, Any]]]]:
    sentences: list[dict[str, Any]] = []
    words: list[str] = []
    labels: list[str] = []
    links: list[str] = []
    id = ""

    with open(file, mode="r", encoding="utf-8") as f:
        for line in f:
            line = line.rstrip()
            if line.startswith("-DOCSTART-"):
                if sentences:
                    yield id, sentences
                    sentences = []
            elif line.startswith("# "):
                id = line[2:].strip().split('\t')[0]
            elif not line:
                if words:
                    sentences.append(_conll_to_example(words, labels, links))
                    words = []
                    labels = []
                    links = []
            else:
                cols = line.split(delimiter)
                words.append(cols[word_column])
                labels.append(cols[tag_column])
                links.append(cols[link_column])

    if sentences:
        yield id, sentences


def _conll_to_example(words: list[str], tags: list[str], links: list[str]) -> dict[str, Any]:
    text, positions = _conll_words_to_text(words)
    entities = [
        {"start": positions[start][0], "end": positions[end - 1][1], "label": [label], "title": [title], 'text': text[positions[start][0]: positions[end - 1][1]]}
        for start, end, label, title in _conll_tags_to_spans(tags, links)
    ]
    return {"text": text, "entities": entities}


def _conll_words_to_text(words: Iterable[str]) -> tuple[str, list[tuple[int, int]]]:
    text = ""
    positions = []
    offset = 0
    for word in words:
        if text:
            text += " "
            offset += 1
        text += word
        n = len(word)
        positions.append((offset, offset + n))
        offset += n
    return text, positions


def _conll_tags_to_spans(tags: Iterable[str], links: Iterable[str]) -> Iterable[tuple[int, int, str, str]]:
    # NOTE: assume BIO scheme
    start, label, link = -1, None, None
    for i, (tag, link_tag) in enumerate(zip(list(tags) + ["O"], list(links) + ["O"])):
        if tag == "O":
            if start >= 0:
                assert label is not None and link is not None
                yield (start, i, label, link)
                start, label, link = -1, None, None
        else:
            cur_label = tag[2:]
            cur_link = link_tag[2:] if tag[:2] == link_tag[:2] else link_tag
            if tag.startswith("B"):
                if start >= 0:
                    assert label is not None and link is not None
                    yield (start, i, label, link)
                start, label, link = i, cur_label, cur_link
            else:
                if cur_label != label:
                    if start >= 0:
                        assert label is not None and link is not None
                        yield (start, i, label, link)
                    start, label, link = i, cur_label, cur_link


def read_text_file(file_path: str) -> str:
    with open(file_path, 'r', encoding='utf-8') as f:
        text = f.read()
    return text


def read_jsonl(file_path: str) -> Iterator[dict[str, Any]]:
    with open(file_path, 'r', encoding='utf-8') as f:
        for line in f:
            yield json.loads(line)



import itertools
import json
import os
from argparse import ArgumentParser
from typing import Any, Iterable, Union

from datasets import DownloadManager

from appendix.data_convert.utils import WikiMapper, _conll_to_example

_URL = "http://resources.mpi-inf.mpg.de/yago-naga/aida/download/KORE50.tar.gz"


def read_conll(
        file: Union[str, bytes, os.PathLike],
        delimiter: str = ' ',
        word_column: int = 0,
        tag_column: int = 1,
        link_column: int = 2
    ) -> Iterable[list[dict[str, Any]]]:
    sentences: list[dict[str, Any]] = []
    words: list[str] = []
    labels: list[str] = []
    links: list[str] = []

    with open(file, mode="r", encoding="utf-8") as f:
        for line in f:
            line = line.rstrip()
            if line.startswith("-DOCSTART-"):
                if sentences:
                    yield sentences
                    sentences = []
            elif not line:
                if words:
                    sentences.append(_conll_to_example(words, labels, links))
                    words = []
                    labels = []
                    links = []
            else:
                cols = line.split(delimiter)
                if len(cols) == 1:
                    words.append(cols[word_column])
                    labels.append("O")
                    links.append("")
                else:
                    words.append(cols[word_column])
                    labels.append(cols[tag_column])
                    links.append(cols[link_column])
        if words:
            sentences.append(_conll_to_example(words, labels, links))
        if sentences:
            yield sentences


def main(finder: WikiMapper, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'kore50')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    data_dir = DownloadManager().download_and_extract(_URL)
    data_file = os.path.join(data_dir, "KORE50", "AIDA.tsv")

    with open(os.path.join(output_dir_data, "test.jsonl"), 'w', encoding='utf-8') as f:
        uid = map(str, itertools.count(start=0, step=1))
        for sentences in read_conll(data_file, delimiter='\t', tag_column=1, link_column=3):
            for example in sentences:
                entities = []
                for ent in example['entities']:
                    title  = ent['title'][0]
                    if title == '--NME--':
                        entities.append({"start": ent['start'], "end": ent['end'], "label": ["-1"]})
                        continue
                    res = finder.find_by_title(title)
                    page_id = res["page_id"] if res is not None and res["page_id"] is not None else "-1"
                    entities.append({
                        "start": ent['start'],
                        "end": ent['end'],
                        "label": [str(page_id)],
                    })

                example = {
                    "subset": "kore50",
                    "id": next(uid),
                    "text": example['text'],
                    "entities": entities
                }
                f.write(f"{json.dumps(example, ensure_ascii=False)}\n")


if __name__ == "__main__":
    parser = ArgumentParser(description="KORE50 Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(finder, output_dir=args.output_dir)
    finder.close()



import random
from typing import Any, Iterator, Optional

from datasets import Dataset
from transformers import BatchEncoding

from entity_linkings.data_utils import (
    EntityDictionary,
    Preprocessor,
    cut_context_window,
    preprocess,
    truncate_around_mention,
)


def process_candidates(
        candidate_titles: list[str],
        gold_titles: list[str],
        separator: str = '*'
    )-> tuple[str, list[int], list[int], list[tuple[int, int]]]:
    context = ""
    candidates_offsets = []
    answer_starts, answer_ends = [], []
    for candidate_title in candidate_titles:
        candidate_start = len(context)
        candidate_end = candidate_start + len(candidate_title)
        candidates_offsets.append( (candidate_start, candidate_end) )
        if candidate_title in gold_titles:
            answer_starts.append(candidate_start)
            answer_ends.append(candidate_end)
        context += candidate_title + f" {separator} "
    return context, answer_starts, answer_ends, candidates_offsets


def compute_char_to_tokens(context: str, context_mask: list[bool], offsets_map: list[tuple[int, int]]) -> dict[int, int]:
    char2token = {}
    first = True
    for _t_idx, (m, cp) in enumerate(zip(context_mask, offsets_map)):
        if m:
            while (
                offsets_map[_t_idx][0] < offsets_map[_t_idx][1]
                and context[offsets_map[_t_idx][0]] == " "
            ):
                offsets_map[_t_idx][0] += 1

            # add prefix space seems to be bugged on some tokenizers
            if first:
                first = False
                if cp[0] != 0 and context[cp[0] - 1] != " ":
                    offsets_map[_t_idx][0] -= 1
                    cp = (cp[0] - 1, cp[1])
            if cp[0] == cp[1]:
                assert context[cp[0] - 1] == " ", f"Token {_t_idx} found to occur at char span ({cp[0]}, {cp[1]}), which is impossible"
            for c in range(*cp):
                char2token[c] = _t_idx
    return char2token


class ExtendPreprocessor(Preprocessor):
    def __init__(self, dictionary: EntityDictionary, candidate_separator: str = "*", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.dictionary = dictionary
        self.candidate_separator = candidate_separator

    def process_context(
            self, text: str, start: int, end: int,
            candidate_ids: Optional[list[str]] = None, labels: Optional[list[str]] = None,
            train: bool = False
        ) -> BatchEncoding:
        if candidate_ids is None:
            raise ValueError("candidate_ids must be provided for ExtendPreprocessor.")

        context, new_start, new_end = cut_context_window(text, start, end, self.context_window_chars)
        marked_text = context[:new_start] + self.ent_start_token + context[new_start:new_end] + self.ent_end_token + context[new_end:]
        input_ids = self.tokenizer.encode(marked_text, add_special_tokens=False, truncation=False)

        start_marker_idx = input_ids.index(self.start_marker_id)
        end_marker_idx = input_ids.index(self.end_marker_id)

        trunscated_input_ids, _, _ = truncate_around_mention(input_ids, self.max_context_length, start_marker_idx, end_marker_idx + 1)

        if labels is not None:
            gold_titles = [self.dictionary(labels[0])['name']]
            if train:
                candidate_ids = candidate_ids[:-1] + [labels[0]]
                random.shuffle(candidate_ids)
        else:
            gold_titles = []

        candidate_titles = [self.dictionary(cand)['name'] for cand in candidate_ids]
        candidate_context, answer_starts, answer_ends, candidates_offsets = process_candidates(
            candidate_titles, gold_titles, separator=self.candidate_separator
        )

        encodings = self.tokenizer(
            self.tokenizer.decode(trunscated_input_ids), candidate_context,
            return_offsets_mapping=True,
        )
        char2token = compute_char_to_tokens(
            candidate_context,
            [p == 1 for p in encodings.sequence_ids()],
            encodings["offset_mapping"]
        )

        encodings["candidates"] = candidate_ids
        encodings["candidates_offsets"] = [
            (char2token[si], char2token[ei - 1] + 1)
            for si, ei in candidates_offsets
        ]
        if labels is not None:
            if train:
                encodings["start_positions"] = [char2token[ans_start] for ans_start in answer_starts]
                encodings["end_positions"] = [char2token[ans_end - 1] + 1 for ans_end in answer_ends]
            encodings["labels"] = labels

        return encodings

    def dataset_preprocess(self, dataset: Dataset, candidates_ids: Optional[list[list[str]]] = None, train: bool = False) -> Dataset:
        def _preprocess_example(examples: Dataset) -> Iterator[BatchEncoding]:
            for i, (text, entity) in enumerate(zip(examples["text"], examples["entity"])):
                candidate_ids = examples["candidates"][i] if "candidates" in examples else None
                encodings = self.process_context(
                    text, entity['start'], entity['end'], candidate_ids,
                    entity['label'] if 'label' in entity else None, train=train
                )
                yield encodings
        flatten_dataset = self.data_flatten(dataset)
        if candidates_ids is not None:
            assert len(flatten_dataset) == len(candidates_ids)
            flatten_dataset = flatten_dataset.add_column("candidates", candidates_ids)
        processed_dataset = preprocess(flatten_dataset, _preprocess_example, desc="Preprocessing")
        return processed_dataset



from importlib.resources import files

from datasets import load_dataset
from transformers import AutoTokenizer

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary

from .preprocessor import FusionedPreprocessor

MODELS = ["google/flan-t5-base"]
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)
candidates_ids = retriever.retrieve_candidates(dataset, top_k=3, only_negative=True)


class TestFusionedPreprocessor:
    def test_init(self) -> None:
        model_name = MODELS[0]
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        preprocessor = FusionedPreprocessor(
            tokenizer=tokenizer,
            ent_start_token="<extra_id_6>",
            ent_end_token="<extra_id_7>",
            title_token="<extra_id_2>",
            entity_token="<extra_id_4>",
        )
        assert preprocessor.title_token == "<extra_id_2>"
        assert preprocessor.ent_start_token == "<extra_id_6>"
        assert preprocessor.ent_end_token == "<extra_id_7>"
        assert preprocessor.entity_token == "<extra_id_4>"
        assert len(preprocessor.prefix_ids) == 0
        assert len(preprocessor.suffix_ids) == 1
        assert preprocessor.max_context_length == 249 # 250 - len(preprocessor.prefix_ids) - len(preprocessor.suffix_ids)
        assert preprocessor.offset_correction == len(preprocessor.prefix_ids)
        assert preprocessor.max_candidate_length == 100
        assert preprocessor.context_window_chars == 1000

    def test_dictionary_preprocess(self) -> None:
        tokenizer = AutoTokenizer.from_pretrained(MODELS[0])
        preprocessor = FusionedPreprocessor(
            tokenizer=tokenizer,
            ent_start_token="<extra_id_6>",
            ent_end_token="<extra_id_7>",
            title_token="<extra_id_2>",
            entity_token="<extra_id_4>",
        )
        processed_dictionary = preprocessor.dictionary_preprocess(dictionary)
        for entry in processed_dictionary:
            assert "encoding" in entry
            assert isinstance(entry["encoding"], list)
            tokens = tokenizer.convert_ids_to_tokens(entry["encoding"])
            assert tokens[0] == "<extra_id_2>"
            assert tokenizer.decode(entry["encoding"]) == f"<extra_id_2> {entry['name']}<extra_id_4> {entry['description']}"
            assert len(entry["encoding"]) <= 32


    def test_process_context(self) -> None:
        tokenizer = AutoTokenizer.from_pretrained(MODELS[0])
        preprocessor = FusionedPreprocessor(
            tokenizer=tokenizer,
            ent_start_token="<extra_id_6>",
            ent_end_token="<extra_id_7>",
            title_token="<extra_id_2>",
            entity_token="<extra_id_4>",
        )
        text = dataset['text'][0]
        entity = dataset['entities'][0][0]
        encodings = preprocessor.process_context(text, entity['start'], entity['end'])
        assert "input_ids" in encodings
        assert "attention_mask" in encodings
        assert "start_positions" not in encodings
        assert "end_positions" not in encodings
        start_pos = encodings['input_ids'].index(preprocessor.start_marker_id)
        end_pos = encodings['input_ids'].index(preprocessor.end_marker_id)
        mention_ids = encodings['input_ids'][start_pos: end_pos + 1]
        mention_text = text[entity['start']: entity['end']]
        decoded_mention = tokenizer.decode(mention_ids, skip_special_tokens=False).replace(" ", "")
        expected_mention = f"{preprocessor.ent_start_token}{mention_text}{preprocessor.ent_end_token}".replace(" ", "")
        assert decoded_mention == expected_mention
        assert len(encodings["input_ids"]) == len(encodings["attention_mask"])
        assert len(encodings["input_ids"]) <= preprocessor.max_context_length




import tempfile
from importlib.resources import files

import pytest
from datasets import load_dataset
from transformers.trainer_utils import TrainOutput

import assets as test_data
from entity_linkings import get_rerankers, get_retrievers, load_dictionary
from entity_linkings.trainer import TrainingArguments
from entity_linkings.utils import BaseSystemOutput

from .crossencoder import CROSSENCODER

MODELS = ["google-bert/bert-base-uncased"]
dataset_path = str(files(test_data).joinpath("dataset_toy_wo_candidates.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)


@pytest.mark.reranker_crossencoder
class TestCROSSENCODER:
    def test__init__(self) -> None:
        model_cls = get_rerankers("crossencoder")
        model = model_cls(
            retriever=retriever,
            config=model_cls.Config(
                model_name_or_path=MODELS[0]
            )
        )
        assert isinstance(model, CROSSENCODER)
        assert hasattr(model, "config") and hasattr(model, "tokenizer") and hasattr(model, "dictionary")
        assert model.config.model_name_or_path == "google-bert/bert-base-uncased"
        assert model.config.nil_token == "[NIL]" and model.config.nil_token in model.tokenizer.all_special_tokens
        assert model.config.ent_start_token == "[START_ENT]" and model.config.ent_start_token in model.tokenizer.all_special_tokens
        assert model.config.ent_end_token == "[END_ENT]" and model.config.ent_end_token in model.tokenizer.all_special_tokens
        assert model.config.entity_token == "[ENT]" and model.config.entity_token in model.tokenizer.all_special_tokens
        assert model.config.max_context_length == 128
        assert model.config.max_candidate_length == 50
        assert model.config.context_window_chars == 500
        assert model.config.pooling == 'first'

    def test_train(self) -> None:
        model = CROSSENCODER(retriever=retriever, config=CROSSENCODER.Config(model_name_or_path=MODELS[0]))
        with tempfile.TemporaryDirectory() as tmpdir:
            result = model.train(
                    train_dataset=dataset,
                    eval_dataset=dataset,
                    num_candidates=3,
                    training_args=TrainingArguments(
                        output_dir=tmpdir,
                        num_train_epochs=1,
                        per_device_train_batch_size=2,
                        per_device_eval_batch_size=2,
                        logging_strategy="no",
                        save_strategy="no",
                        eval_strategy="no",
                        remove_unused_columns=False,
                        eval_on_start=True
                    )
            )
            assert isinstance(result, TrainOutput)
            assert hasattr(result, 'metrics')

    def test_evaluate(self) -> None:
        model = CROSSENCODER(retriever=retriever, config=CROSSENCODER.Config(model_name_or_path=MODELS[0]))
        metrics = model.evaluate(dataset, num_candidates=3, batch_size=1)
        assert 'top1_accuracy' in metrics

    def test_predict(self) -> None:
        sentence = "Apple is looking at buying U.K. startup for $1 billion"
        spans = [(0,5), (27,30), (44,52)]
        model = CROSSENCODER(retriever=retriever, config=CROSSENCODER.Config(model_name_or_path=MODELS[0]))
        predictions = model.predict(sentence, spans=spans, num_candidates=3)
        assert isinstance(predictions, list)
        assert len(predictions) == len(spans)
        for preds in predictions:
            assert isinstance(preds, BaseSystemOutput)
            assert preds.id is not None and preds.start is not None and preds.end is not None




import logging
from dataclasses import dataclass
from typing import Any, Optional, Union

import torch
import torch.nn as nn
import transformers
from transformers import Trainer
from transformers.trainer_pt_utils import nested_detach

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@dataclass
class TrainingArguments(transformers.TrainingArguments):
    remove_unused_columns: bool = False
    lr_scheduler_type: str = "linear"
    warmup_ratio: float = 0.1
    optim: str = "adamw_torch"
    adam_beta1: float = 0.9
    adam_beta2: float = 0.98
    adam_epsilon: float = 1.e-6
    learning_rate: float = 1.e-5
    weight_decay: float = 0.01
    max_grad_norm: float = 0.0
    log_level: str = "info"
    logging_strategy: str = "steps"
    logging_steps: int = 500
    save_strategy: str = "epoch"
    save_total_limit: int = 3
    eval_strategy: str = "epoch"
    metric_for_best_model: str = "loss"
    load_best_model_at_end: bool = True
    eval_on_start: bool = False


class EntityLinkingTrainer(Trainer):
    def compute_loss(
        self,
        model: nn.Module,
        inputs: dict[str, torch.Tensor],
        return_outputs: bool = False,
        num_items_in_batch: Optional[torch.Tensor] = None,
    ) -> Union[torch.Tensor, tuple[torch.Tensor, Any]]:
        """
        How the loss is computed by Trainer. By default, all models return the loss in the first element.

        Args:
            model (`nn.Module`):
                The model to compute the loss for.
            inputs (`dict[str, Union[torch.Tensor, Any]]`):
                The input data for the model.
            return_outputs (`bool`, *optional*, defaults to `False`):
                Whether to return the model outputs along with the loss.
            num_items_in_batch (Optional[torch.Tensor], *optional*):
                The number of items in the batch. If num_items_in_batch is not passed,

        Returns:
            The loss of the model along with its output if return_outputs was set to True

        Subclass and override for custom behavior. If you are not using `num_items_in_batch` when computing your loss,
        make sure to overwrite `self.model_accepts_loss_kwargs` to `False`. Otherwise, the loss calculating might be slightly inaccurate when performing gradient accumulation.
        """

        if (self.label_smoother is not None or self.compute_loss_func is not None) and "labels" in inputs:
            labels = inputs.pop("labels")
        else:
            labels = None

        # if self.model_accepts_loss_kwargs:
            # kwargs = {}
            # if num_items_in_batch is not None:
            #     kwargs["num_items_in_batch"] = num_items_in_batch
            # inputs = {**inputs, **kwargs}

        # User-defined compute_loss function
        if self.compute_loss_func is not None:
            if labels is None:
                logger.warning(
                    "Trainer: `compute_loss_func` is defined but `labels=None`. "
                    "Your custom loss function will still be called with labels=None. "
                )
            loss, outputs = self.compute_loss_func(inputs, labels)
        else:
            outputs = model(**inputs)
            if isinstance(outputs, dict) and "loss" not in outputs:
                raise ValueError(
                    "The model did not return a loss from the inputs, only the following keys: "
                    f"{','.join(outputs.keys())}. For reference, the inputs it received are {','.join(inputs.keys())}."
                )
            loss = outputs["loss"] if isinstance(outputs, dict) else outputs[0]

        # Save past state if it exists
        # TODO: this needs to be fixed and made cleaner later.
        if self.args.past_index >= 0:
            self._past = outputs[self.args.past_index]

        if self.args.average_tokens_across_devices and num_items_in_batch is not None:
            loss *= self.accelerator.num_processes if self.args.n_gpu <= 1 else self.args.n_gpu

        return (loss, outputs) if return_outputs else loss

    def save_model(self, output_dir: Optional[str] = None, _internal_call: bool = False) -> None:
        if output_dir is None:
            output_dir = self.args.output_dir
        self.model.save_pretrained(output_dir)

    def prediction_step(
        self,
        model: nn.Module,
        inputs: dict[str, Union[torch.Tensor, Any]],
        prediction_loss_only: bool,
        ignore_keys: Optional[list[str]] = None,
    ) -> tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]:
        """
        Perform an evaluation step on `model` using `inputs`.

        Subclass and override to inject custom behavior.

        Args:
            model (`nn.Module`):
                The model to evaluate.
            inputs (`dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.

                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument `labels`. Check your model's documentation for all accepted arguments.
            prediction_loss_only (`bool`):
                Whether or not to return the loss only.
            ignore_keys (`list[str]`, *optional*):
                A list of keys in the output of your model (if it is a dictionary) that should be ignored when
                gathering predictions.

        Return:
            Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]: A tuple with the loss,
            logits and labels (each being optional).
        """
        has_labels = False if len(self.label_names) == 0 else all(inputs.get(k) is not None for k in self.label_names)

        inputs = self._prepare_inputs(inputs)
        if ignore_keys is None:
            if hasattr(self.model, "config"):
                ignore_keys = getattr(self.model.config, "keys_to_ignore_at_inference", ["past_key_values"])
            else:
                ignore_keys = []

        # labels may be popped when computing the loss (label smoothing for instance) so we grab them first.
        if has_labels:
            labels = nested_detach(tuple(inputs.get(name) for name in self.label_names))
            if len(labels) == 1:
                labels = labels[0]
        else:
            labels = None

        with torch.no_grad():
            if has_labels:
                with self.compute_loss_context_manager():
                    num_items_in_batch = self._get_num_items_in_batch([inputs], self.args.device)
                    loss, outputs = self.compute_loss(
                        model, inputs, return_outputs=True, num_items_in_batch=num_items_in_batch
                    )
                loss = loss.detach().mean()

                if isinstance(outputs, dict):
                    logits = tuple(v for k, v in outputs.items() if k not in ignore_keys + ["loss"])
                else:
                    logits = outputs[1:]
            else:
                loss = None
                with self.compute_loss_context_manager():
                    outputs = model(**inputs)
                if isinstance(outputs, dict):
                    logits = tuple(v for k, v in outputs.items() if k not in ignore_keys)
                else:
                    logits = outputs
                # TODO: this needs to be fixed and made cleaner later.
                if self.args.past_index >= 0:
                    self._past = outputs[self.args.past_index - 1]

        if prediction_loss_only:
            return (loss, None, None)

        detached_logits = nested_detach(logits)
        if len(detached_logits) == 1:
            detached_logits = detached_logits[0]

        return (loss, detached_logits, labels)



from typing import Any, Optional

from transformers import BatchEncoding

from entity_linkings.data_utils import (
    EntityDictionary,
    Preprocessor,
    cut_context_window,
    truncate_around_mention,
)


class FusionedPreprocessor(Preprocessor):
    def __init__(self, title_token: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.title_token = title_token

    def dictionary_preprocess(self, dictionary: EntityDictionary) -> EntityDictionary:
        def preprocess_example(name: str, description: str) -> dict[str, list[int]]:
            text = self.title_token + name + self.entity_token + description
            encodings  = self.tokenizer.encode(
                text,
                padding=True,
                truncation=True,
                add_special_tokens=False,
                max_length=self.max_candidate_length,
            )
            return encodings
        dictionary.add_encoding(preprocess_example)
        return dictionary

    def process_context(self, text: str, start: int, end: int, candidate_ids: Optional[list[str]] = None, labels: Optional[list[str]] = None) -> BatchEncoding:
        context, new_start, new_end = cut_context_window(text, start, end, self.context_window_chars)
        marked_text = context[:new_start] + self.ent_start_token + context[new_start:new_end] + self.ent_end_token + context[new_end:]
        input_ids = self.tokenizer.encode(marked_text, add_special_tokens=False, truncation=False)

        start_marker_idx = input_ids.index(self.start_marker_id)
        end_marker_idx = input_ids.index(self.end_marker_id)

        truncated_input_ids, _, _ = truncate_around_mention(input_ids, self.max_context_length, start_marker_idx, end_marker_idx + 1)
        encodings = BatchEncoding({
            "input_ids": truncated_input_ids,
            "attention_mask": [1] * len(truncated_input_ids),
        })
        if labels is not None:
            encodings["labels"] = labels
        if candidate_ids is not None:
            encodings["candidates"] = candidate_ids
        return encodings



from dataclasses import dataclass
from typing import Any, Callable

import torch
import yaml
from transformers import AutoConfig, AutoModel, PreTrainedModel, T5EncoderModel

ENCODER_DECODER_MODELS = {
    "T5Config": T5EncoderModel,
}


@dataclass
class BaseSystemOutput:
    query: str
    start: int
    end: int
    id: str


def read_yaml(path: str) -> dict[str, Any]:
    with open(path, 'r') as yml:
        config = yaml.safe_load(yml)
    return config


def load_model(model_name_or_path: str) -> PreTrainedModel:
    model_config = AutoConfig.from_pretrained(model_name_or_path)

    if type(model_config).__name__ in ENCODER_DECODER_MODELS:
        model_cls = ENCODER_DECODER_MODELS.get(type(model_config).__name__, None)
        if not model_cls:
            raise NotImplementedError(f"Encoder-decoder models are not supported. {model_config}")
        model = model_cls.from_pretrained(model_name_or_path, config=model_config)
    else:
        if model_config.is_encoder_decoder:
            raise NotImplementedError(f"Encoder-decoder models are not supported. {model_config}")
        model = AutoModel.from_pretrained(model_name_or_path, config=model_config)
    return model


def get_pooler(pooling: str) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """
    Get the pooling function based on the specified method.
    Args:
        - pooling (`str`):
            The pooling method to use. Options are 'first', 'mean', or 'last'.
    Returns:
        - `Callable[[torch.Tensor, torch.Tensor], torch.Tensor]`: The pooling function.
    """
    if pooling == 'first':
        return first_token_pooler
    elif pooling == 'mean':
        return mean_tokens_pooler
    elif pooling == 'last':
        return last_token_pooler
    elif pooling == 'concat':
        return concat_pooler
    else:
        raise ValueError(f"Invalid pooling method: {pooling}. Choose either 'first', 'mean', or 'last'.")


def first_token_pooler(last_hidden_states: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
    """
    Pooling using the first token ([CLS] or [BOS]) representation.
    Args:
        - last_hidden_states (`torch.Tensor`):
            The last hidden states from the model of shape (batch_size, sequence_length, hidden_size).
        - attention_mask (`torch.Tensor`):
            The attention mask of shape (batch_size, sequence_length), where 1 indicates valid tokens
    Returns:
        - `torch.Tensor`: The pooled output of shape (batch_size, hidden_size).
    """
    if attention_mask[:, 0].sum() == 0 and attention_mask[:, -1].sum() > 0:
        # left padding is used, use the first non-padding token as the last token.
        seq_lens = attention_mask.sum(dim=1)
        return last_hidden_states[:, -seq_lens, :]
    else:
        # right padding is used. use the first token as the start token.
        return last_hidden_states[:, 0, :]


def mean_tokens_pooler(last_hidden_states: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
    """
    Pooling by averaging the token representations, weighted by the attention mask.
    Args:
        - last_hidden_states (`torch.Tensor`):
            The last hidden states from the model of shape (batch_size, sequence_length, hidden_size).
        - attention_mask (`torch.Tensor`):
            The attention mask of shape (batch_size, sequence_length), where 1 indicates valid tokens and 0 indicates padding.
    Returns:
        - `torch.Tensor`: The pooled output of shape (batch_size, hidden_size).
    """
    last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)
    return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]


def last_token_pooler(last_hidden_states: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
    """
    Pooling using the last token representation.
    Args:
        - last_hidden_states (`torch.Tensor`):
            The last hidden states from the model of shape (batch_size, sequence_length, hidden_size).
        - attention_mask (`torch.Tensor`):
            The attention mask of shape (batch_size, sequence_length), where 1 indicates valid tokens and 0 indicates padding.
    Returns:
        - `torch.Tensor`: The pooled output of shape (batch_size, hidden_size).
    """
    if attention_mask[:, 0].sum() == 0 and attention_mask[:, -1].sum() > 0:
        # left padding is used, use the last token as the last token.
        return last_hidden_states[:, -1, :]
    else:
        # right padding is used. use the last non-padding token as the last token.
        seq_lens = attention_mask.sum(dim=1) - 1
        return last_hidden_states[torch.arange(last_hidden_states.size(0)), seq_lens, :]


def concat_pooler(last_hidden_states: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
    """
    Pooling by concatenating the first and last token representations.
    Args:
        - last_hidden_states (`torch.Tensor`):
            The last hidden states from the model of shape (batch_size, sequence_length, hidden_size).
    Returns:
        - `torch.Tensor`: The pooled output of shape (batch_size, 2 * hidden_size).
    """
    first_token = first_token_pooler(last_hidden_states, attention_mask)
    last_token = last_token_pooler(last_hidden_states, attention_mask)
    return torch.cat((first_token, last_token), dim=1)


def calculate_top1_accuracy(num_corrects: int, num_golds: int) -> dict[str, float]:
    return {
        "top1_accuracy": num_corrects / num_golds if num_golds > 0 else 0.
    }


def calculate_recall_mrr(predictions: list[dict[str, Any]]) -> dict[str, int | float]:
    true, tp_1, tp_10, tp_50, tp_100, reciprocal_rank = 0, 0, 0, 0, 0, 0.
    for prediction in predictions:
        true += 1
        best_rank = 0
        indices = [result['id'] for result in prediction['predict']]
        for label in prediction['gold']:
            if label in indices:
                rank = indices.index(label) + 1
                if rank < best_rank or best_rank == 0:
                    best_rank = rank
        if best_rank > 0:
            if best_rank == 1:
                tp_1 += 1
            if best_rank <= 10:
                tp_10 += 1
            if best_rank <= 50:
                tp_50 += 1
            if best_rank <= 100:
                tp_100 += 1
            reciprocal_rank += 1 / best_rank

    return {
        "recall@1": tp_1 / true if true > 0 else 0.,
        "recall@10": tp_10 / true if true > 0 else 0.,
        "recall@50": tp_50 / true if true > 0 else 0.,
        "recall@100": tp_100 / true if true > 0 else 0.,
        "mrr@100": reciprocal_rank / true if true > 0. else 0.
    }


def calculate_inkb_f1(predictions: list[list[dict[str, Any]]], golds: list[list[dict[str, Any]]]) -> dict[str, int | float]:
    tp, t, p = 0, 0, 0
    assert len(predictions) == len(golds)
    for prediction, gold in zip(predictions, golds):
        gold_ids = {(gold['start'], gold['end']): gold['label'] for gold in gold}
        pred_ids = {(result['start'], result['end']): result['label'] for result in prediction}
        for p_span, p_label in pred_ids.items():
            if p_span in gold_ids and len(set(p_label) & set(gold_ids[p_span])) > 0:
                tp += 1
        t += len(gold_ids)
        p += len(pred_ids)

    precision = tp / p if p > 0 else 0.
    recall = tp / t if t > 0 else 0.
    f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.

    return {
        "precision": precision,
        "recall": recall,
        "f1": f1
    }



from importlib.resources import files

import assets as test_data
from entity_linkings import load_dictionary

from .base import IndexerBase, RetrieverBase

dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)


def test_IndexerBase() -> None:
    indexer = IndexerBase(dictionary=dictionary)
    assert isinstance(indexer, IndexerBase)
    assert indexer.dictionary == dictionary

def test_RetrieverBase() -> None:
    linker = RetrieverBase(dictionary=dictionary)
    assert isinstance(linker, RetrieverBase)
    assert hasattr(linker, 'config')
    assert isinstance(linker.config, RetrieverBase.Config)



import abc
import os
from dataclasses import dataclass
from typing import Any, Callable, Iterator, Optional, TypedDict

import datasets
from datasets import Dataset
from transformers import TrainingArguments


class Entity(TypedDict):
    id: str
    name: str
    description: str
    label_id: int
    encoding: Optional[dict[str, list[int]]]


@dataclass
class EntityDictionaryConfig(datasets.BuilderConfig):
    """BuilderConfig for EntityDictionary."""
    name: Optional[str] = None
    version: Optional[datasets.Version] = None
    description: Optional[str] = None


class EntityDictionary(abc.ABC):
    def __init__(
            self,
            dictionary: Dataset,
            nil_id: str = "-1",
            nil_name: str = "<NIL>",
            nil_description: str = "<NIL> is an entity that does not exist in this dictionary.",
            default_description: str = """{name} is an entity in this dictionary.""",
            cache_dir: Optional[str|os.PathLike] = None,
        ) -> None:
        self.entity_dict = dictionary
        self.entity_dict = self.entity_dict.add_item({
            "id": nil_id,
            "name": nil_name,
            "description": nil_description
        })
        self.id_to_index = {id: i for i, id in enumerate(self.entity_dict["id"])}
        self.title_to_index = {title: i for i, title in enumerate(self.entity_dict["name"])}
        self.nil_id = nil_id
        self.nil_name = nil_name
        self.nil_description = nil_description
        self.default_description = default_description
        self.cache_dir = cache_dir

    def __call__(self, entity_id: str) -> Entity:
        if entity_id in self.id_to_index:
            index = self.id_to_index[entity_id]
            return self.entity_dict[index]
        else:
            nil_index = self.id_to_index[self.nil_id]
            return self.entity_dict[nil_index]

    def __iter__(self) -> Iterator[Entity]:
        for value in self.entity_dict:
            yield value

    def __getitem__(self, idx: int) -> Entity:
        return self.entity_dict[idx]

    def __len__(self) -> int:
        return len(self.entity_dict)

    def __repr__(self) -> str:
        return self.entity_dict.__repr__()

    def get_label_id(self, entity_id: str) -> int:
        if entity_id in self.id_to_index:
            return self.id_to_index[entity_id]
        else:
            return self.id_to_index[self.nil_id]

    def get_from_title(self, title: str) -> Entity:
        if title in self.title_to_index:
            index = self.title_to_index[title]
            return self.entity_dict[index]
        else:
            nil_index = self.title_to_index[self.nil_name]
            return self.entity_dict[nil_index]

    def get_entity_ids(self) -> list[str]:
        return list(self.entity_dict["id"])

    def get_entity_names(self) -> list[str]:
        return list(self.entity_dict["name"])

    def get_entity_descriptions(self) -> list[str]:
        return list(self.entity_dict["description"])

    def convert_description(self, name: str, description: Optional[str] = None) -> str:
        if description:
            return description
        else:
            return self.default_description.format(name=name)

    def add_encoding(
            self,
            tokenizer_func: Callable[[str, str], dict[str, list[int]]],
            training_arguments: Optional[TrainingArguments] = None
        ) -> None:
        def preprocess(documents: Dataset) -> dict[str, list[Any]]:
            outputs: dict[str, list[Any]] = {"id": [], "name": [], "description": [], "encoding": []}
            for id, name, description in zip(documents["id"], documents["name"], documents["description"]):
                description = self.convert_description(name, description)
                outputs["id"].append(id)
                outputs["name"].append(name)
                outputs["description"].append(description)
                outputs["encoding"].append(tokenizer_func(name, description))
            return outputs

        if training_arguments:
            with training_arguments.main_process_first(desc="dataset map pre-processing"):
                column_names = next(iter(self.entity_dict)).column_names
                dictionary = self.entity_dict.map(preprocess, batched=True, remove_columns=column_names)
        else:
            column_names = self.entity_dict.column_names
            dictionary = self.entity_dict.map(preprocess, batched=True, remove_columns=column_names)

        self.entity_dict = dictionary



import hashlib
import json
import os
import re
from typing import Any

import httpx

DEFAULT_TIMEOUT = httpx.Timeout(150.0, read=150.0, write=50.0, connect=6.0)


def process_multi_choice_prompt(multi_choice_prompt_result: str, candidates: list[str]) -> int:
    L = len(candidates)
    if L == 0:
        return -1
    elif L == 1:
        return 0
    elif 'None of the entity match' in multi_choice_prompt_result:
        return -1

    # update the index finding schema with regular expression.
    index_list = [int(s) - 1 for s in re.findall(r'\b\d+\b', multi_choice_prompt_result) if 0 <= int(s) - 1 < len(candidates)]

    # consider direct index answer of chatgpt
    if len(index_list) == 1:
        return index_list[0]

    # if there are two choices and candidate entities length is more than 2, select the first one.
    if len(index_list) == 2 and len(candidates) > 2:
        return index_list[0]

    # consider complete string match
    index_list = []
    for index, candidate in enumerate(candidates):
        if candidate.lower() in multi_choice_prompt_result.lower():
            add_flag = True
            other_candidates = candidates[:index] + candidates[index+1:]
            for other_candidate in other_candidates:
                if candidate in other_candidate:
                    add_flag = False
                    break
            if add_flag:
                index_list.append(index)

    if len(index_list) == 1:
        return index_list[0]
    return -1


class Cache:
    def __init__(self, model_name: str, cache_dir: str = '.cache') -> None:
        # Save GPT output
        os.makedirs(cache_dir, exist_ok = True)
        self.file_name_cache = f'{cache_dir}/{model_name}.jsonl'
        if os.path.exists(self.file_name_cache):
            # This file contains hash values and raw data
            self.cache = self.load_json()
        else:
            self.cache = dict()

    def serialize(self, obj: Any) -> Any:
        """ Serialize an object to a JSON-serializable format."""
        if isinstance(obj, dict):
            return {k: self.serialize(v) for k, v in obj.items()}
        elif hasattr(obj, "__dict__"):
            return self.serialize(obj.__dict__)
        elif isinstance(obj, list):
            return [self.serialize(v) for v in obj]
        else:
            return obj

    @staticmethod
    def create_hash(prompt: str) -> str:
        # Save on an instruction basis
        return hashlib.md5(prompt.encode()).hexdigest()

    def load_json(self) -> dict[str, Any]:
        data = dict()
        with open(self.file_name_cache) as f:
            json_lines = f.readlines()
            for line in json_lines:
                json_obj = json.loads(line)
                data[json_obj['id']] = json_obj['results']
        return data

    def append_to_jsonl(self, instruction: str, results: dict[str, Any]) -> None:
        # Save data as cache.
        self.cache[self.create_hash(instruction)] = results
        save_data = {'id': self.create_hash(instruction), 'results': results}
        with open(self.file_name_cache, 'a') as file:
            json_str = json.dumps(save_data, ensure_ascii=False)
            file.write(json_str + '\n')

    def __call__(self, item: str) -> dict[str, Any]:
        return self.cache[self.create_hash(item)]

    def check_in_cache(self, instruction: str) -> bool:
        return True if self.create_hash(instruction) in self.cache else False



import os
from argparse import ArgumentParser

import yaml

import entity_linkings
from entity_linkings.trainer import TrainingArguments

training_arguments = [
    # Dataloader
    "remove_unused_columns",

    # Training Parameters
    "lr_scheduler_type",
    "warmup_ratio",

    # Optimizer
    "optim",
    "adam_beta1",
    "adam_beta2",
    "adam_epsilon",

    # Learning Rate and Weight Decay
    "learning_rate",
    "weight_decay",
    "max_grad_norm",

    # Logging
    "log_level",
    "logging_strategy",
    "logging_steps",
    "report_to",

    # Save
    "save_strategy",
    "save_total_limit",

    # Evaluation
    "eval_strategy",
    "metric_for_best_model",
    "load_best_model_at_end",
    "eval_on_start"
]


def write_yaml(output_path: str, content: dict) -> None:
    with open(output_path, 'w') as f:
        yaml.dump(content, f)
        f.write('\n')
        yaml.dump({
            "training_arguments": {
                k: v for k, v in TrainingArguments.__dict__.items()
                if k in training_arguments}
        }, f)


def generate_config(output_dir: str) -> None:
    os.makedirs(output_dir, exist_ok=True)
    retriever_ids = entity_linkings.get_retriever_ids()
    reranker_ids = entity_linkings.get_reranker_ids()

    for model_id in retriever_ids:
        output_path = os.path.join(output_dir, f"{model_id}.yaml")
        model_cls = entity_linkings.get_retrievers(model_id)
        model_config = model_cls.Config()
        write_yaml(output_path, {f"{model_id}": model_config.__dict__})

    for model_id in reranker_ids:
        output_path = os.path.join(output_dir, f"{model_id}.yaml")
        model_cls = entity_linkings.get_rerankers(model_id)
        model_config = model_cls.Config()
        write_yaml(output_path, {f"{model_id}": model_config.__dict__})


def cli_main() -> None:
    parser = ArgumentParser()
    parser.add_argument("--output_dir", "-o", type=str, default="configs", help="Output directory for the generated config files.")
    args = parser.parse_args()
    generate_config(args.output_dir)


if __name__ == "__main__":
    cli_main()



import json
import os
from typing import Optional

import torch
import torch.nn as nn

from entity_linkings.utils import get_pooler, load_model


class Encoder(nn.Module):
    '''
    Cross Encoder for entity disambiguation.
    '''

    # to suppress an AttributeError when training
    _keys_to_ignore_on_save = None
    def __init__(self, model_name_or_path: str, pooling: str = 'first') -> None:
        super().__init__()
        self.encoder = load_model(model_name_or_path)
        self.projection = nn.Linear(self.encoder.config.hidden_size, 1)
        self.pooler = get_pooler(pooling)
        self.config = {"model_name_or_path": model_name_or_path, "pooling": pooling}

    def resize_token_embeddings(self, new_num_tokens: int) -> None:
        self.encoder.resize_token_embeddings(new_num_tokens)

    def encode(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, token_type_ids: Optional[torch.Tensor] = None) -> torch.Tensor:
        model_inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        if token_type_ids is not None:
            model_inputs["token_type_ids"] = token_type_ids
        encoder_outputs = self.encoder(**model_inputs)
        last_hidden_state = encoder_outputs.last_hidden_state
        pooled_output = self.pooler(last_hidden_state, attention_mask)
        return pooled_output # (bs, hidden_size)

    def forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, token_type_ids: Optional[torch.Tensor] = None, labels: Optional[torch.Tensor] = None) -> tuple[torch.Tensor|None, torch.Tensor]:
        bs, cs, length = input_ids.size()
        input_ids = input_ids.view(bs * cs, length)
        attention_mask = attention_mask.view(bs * cs, length)
        if token_type_ids is not None:
            token_type_ids = token_type_ids.view(bs * cs, length)

        outputs = self.encode(input_ids, attention_mask, token_type_ids)  # (bs, hidden_size)
        scores = self.projection(outputs).squeeze(-1)  # (bs, )
        scores = scores.view(bs, cs) # (bs, cs)

        loss = None
        if labels is not None:
            loss = nn.functional.cross_entropy(scores, labels)
        return (loss, scores)

    def save_pretrained(self, save_directory: str) -> None:
        self.encoder.save_pretrained(save_directory)
        torch.save(self.projection.state_dict(), os.path.join(save_directory, "projection.pt"))
        model_config = {"model_name_or_path": self.config["model_name_or_path"], "pooling": self.config["pooling"]}
        json.dump(model_config, open(os.path.join(save_directory, "model_config.json"), "w"), indent=2, ensure_ascii=False)

    @classmethod
    def from_pretrained(cls, load_directory: str) -> "CrossEncoder":
        model_config = json.load(open(os.path.join(load_directory, "model_config.json")))
        model_config["model_name_or_path"] = load_directory
        model = cls(**model_config)
        model.projection.load_state_dict(torch.load(os.path.join(load_directory, "projection.pt")))
        return model



from dataclasses import dataclass
from typing import Any, Optional, Union

import torch
from transformers import BatchEncoding
from transformers.data.data_collator import pad_without_fast_tokenizer_warning

from entity_linkings.data_utils import CollatorBase, EntityDictionary


@dataclass
class CollatorForExtend(CollatorBase):
    dictionary: Optional[EntityDictionary] = None
    modify_global_attention: int = 2

    def __call__(self, features: list[Union[dict[str, Any], BatchEncoding]]) -> BatchEncoding:
        candidates_offsets = []
        starts, ends = [], []
        features = [f.copy() for f in features]
        for f in features:
            _ = f.pop("offset_mapping", None)
            _ = f.pop("labels", None)
            _ = f.pop("candidates", None)
            candidates_offsets.append(f.pop("candidates_offsets"))
            start_positions = f.pop("start_positions", None)
            end_positions = f.pop("end_positions", None)
            if start_positions is not None and end_positions is not None:
                starts.append(start_positions[0])
                ends.append(end_positions[0])

        batch = pad_without_fast_tokenizer_warning(
            self.tokenizer,
            features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors=self.return_tensors,
        )

        global_attention = torch.zeros_like(batch["attention_mask"])
        if self.modify_global_attention == 1:
            global_attention[:, 0] = 1  # CLS global attention
            for i, candidate_offsets in enumerate(candidates_offsets):
                for si, ei in candidate_offsets:
                    global_attention[i, si] = 1
                    global_attention[i, ei] = 1
        elif self.modify_global_attention == 2:
            first_candidate_starts = [
                min([si for si, _ in cand_offs])
                for cand_offs in candidates_offsets
            ]
            for i, fcs in enumerate(first_candidate_starts):
                global_attention[i, :fcs] = 1
        else:
            raise ValueError(f"Unknown modify_global_attention: {self.modify_global_attention}")

        batch["global_attention_mask"] = global_attention if self.return_tensors == "pt" else global_attention.tolist()
        if starts and ends:
            batch['start_positions'] = torch.tensor(starts) if self.return_tensors == "pt" else starts
            batch['end_positions'] = torch.tensor(ends) if self.return_tensors == "pt" else ends

        return batch





import json
import os
from argparse import ArgumentParser

import requests


def convert_zelda_train(train_file: str, output_dir: str) -> None:
    dataset = []
    with open(train_file) as f:
        for line in f:
            line = json.loads(line)
            assert len(line['index']) == len(line['wikipedia_ids'])
            entities = []
            for span, label in zip(line['index'], line['wikipedia_ids']):
                entities.append({
                    "start": span[0],
                    "end": span[1],
                    "label": [str(label)],
                })
            example = {
                "subset": 'zelda',
                "id": str(line['page_id'])+'_'+str(line['section_name']),
                "text": line["text"],
                "entities": entities
            }
            dataset.append(example)

    num_train = int(len(dataset) * 0.9)
    with open(os.path.join(output_dir, "train.jsonl"), 'w', encoding='utf-8') as f:
        for example in dataset[:num_train]:
            f.write(f"{json.dumps(example, ensure_ascii=False)}\n")
    with open(os.path.join(output_dir, "dev.jsonl"), 'w', encoding='utf-8') as f:
        for example in dataset[num_train:]:
            f.write(f"{json.dumps(example, ensure_ascii=False)}\n")


def convert_zelda_test(output_dir: str) -> None:
    test_dataset_names = ['aida-b', 'cweb', 'wned-wiki', 'reddit-comments', 'reddit-posts', 'tweeki', 'shadowlinks-shadow', 'shadowlinks-tail', 'shadowlinks-top']
    with open(os.path.join(output_dir, "test.jsonl"), 'w', encoding='utf-8') as f:
        for test_dataset_name in test_dataset_names:
            url = f"https://github.com/flairNLP/zelda/raw/refs/heads/main/test_data/jsonl/test_{test_dataset_name}.jsonl"
            urlData = requests.get(url, stream=True)
            for chunk in urlData.iter_lines():
                line = json.loads(chunk)
                assert len(line['index']) == len(line['wikipedia_ids'])
                entities = []
                for span, label in zip(line['index'], line['wikipedia_ids']):
                    entities.append({
                        "start": span[0],
                        "end": span[1],
                        "label": [str(label)],
                    })
                example = {
                    "subset": test_dataset_name,
                    "id": str(line['id']),
                    "text": line["text"],
                    "entities": entities
                }
                f.write(f"{json.dumps(example, ensure_ascii=False)}\n")


def main(train_file: str, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'zelda')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    convert_zelda_train(train_file, output_dir_data)
    convert_zelda_test(output_dir_data)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--train_file", '-t', type=str, default="zelda_train.jsonl", help="Path to the zelda training file.")
    parser.add_argument("--dict_file", '-d', type=str, default="zelda_wiki_dictionary.jsonl", help="Path to the zelda wiki dictionary file.")
    parser.add_argument("--output_dir", '-o', type=str, default="processed_data", help="Output directory for processed data.")
    args = parser.parse_args()

    main(args.train_file, args.output_dir)



from importlib.resources import files

import pytest
from datasets import load_dataset
from transformers import AutoTokenizer

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary

from .preprocessor import ExtendPreprocessor, compute_char_to_tokens, process_candidates

MODEL = "allenai/longformer-base-4096"
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
tokenizer = AutoTokenizer.from_pretrained(MODEL)
tokenizer.add_special_tokens({'additional_special_tokens': ['[START_ENT]', '[END_ENT]', '[NIL]']})
assert tokenizer.is_fast

preprocessor = ExtendPreprocessor(
    tokenizer=tokenizer,
    dictionary=dictionary,
    ent_start_token="[START_ENT]",
    ent_end_token="[END_ENT]",
    entity_token="[ENT]",
    max_context_length=512,
    max_candidate_length=128,
)
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)
sentence = "Apple is looking at buying U.K. startup for $1 billion"
spans = [(0,5), (27,30), (44,52)]
candidates = retriever.predict(sentence, spans=spans, top_k=3)
candidate_ids = [[cand.id for cand in candidates_i] for candidates_i in candidates]

@pytest.mark.parametrize("labels", [["000013"], None])
def test_process_context(labels: list[str]| None) -> None:
    for (b, e), candidates_i in zip(spans, candidate_ids):
        encodings = preprocessor.process_context(sentence, b, e, candidates_i, labels)
        assert "input_ids" in encodings
        assert "attention_mask" in encodings
        assert "offset_mapping" in encodings
        assert "candidates_offsets" in encodings
        if labels is not None:
            assert "labels" in encodings and encodings["labels"]

def test_dataset_preprocess() -> None:
    dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
    dataset = load_dataset("json", data_files={"test": dataset_path})['test']
    candidates = retriever.retrieve_candidates(dataset, top_k=3, only_negative=True)
    processed_dataset = preprocessor.dataset_preprocess(dataset, candidates)
    assert len(processed_dataset) == 8
    for processed in processed_dataset:
        assert "input_ids" in processed
        assert "attention_mask" in processed
        assert "candidates_offsets" in processed
        assert "labels" in processed and processed["labels"]


@pytest.mark.parametrize("gold_titles", [["Meta"], []])
def test_process_candidates(gold_titles: list[str]) -> None:
    candidate_titles = ["Apple", "Meta", "Amazon"]
    gold_titles = ["Meta", "Google"]
    context, answer_starts, answer_ends, candidates_offsets = process_candidates(
        candidate_titles,
        gold_titles,
        separator='*'
    )
    assert context == "Apple * Meta * Amazon * "
    assert answer_starts == [8] if gold_titles else []
    assert answer_ends == [12] if gold_titles else []
    assert candidates_offsets == [(0, 5), (8, 12), (15, 21)]


def test_compute_char_to_tokens() -> None:
    text = "Steve Jobs was found [START_ENT]Apple[END_ENT]."
    candidate_text = "Apple * Meta * Amazon * Google *"

    encodings = tokenizer(
        text,
        candidate_text,
        return_offsets_mapping=True,
    )

    char2token = compute_char_to_tokens(
        candidate_text,
        [p == 1 for p in encodings.sequence_ids()],
        encodings['offset_mapping']
    )
    assert isinstance(char2token, dict)



import logging
from typing import Any, Callable, Iterator, Optional

from datasets import Dataset
from transformers import BatchEncoding, PreTrainedTokenizer, TrainingArguments

from .entity_dictionary import EntityDictionary
from .utils import cut_context_window, truncate_around_mention

logger = logging.getLogger(__name__)


def preprocess(
        dataset: Dataset,
        processing_func: Callable[[Dataset], Iterator[Any]],
        training_arguments: Optional[TrainingArguments] =None,
        desc: Optional[str] = None
        ) -> dict[str, Dataset]:
    def _preprocess(documents: Dataset) -> Any:
        features = [_ for _ in processing_func(documents)]
        outputs = {}
        for k in list(features[0].keys()):
            outputs[k] = [f[k] for f in features]
        return outputs

    if training_arguments:
        with training_arguments.main_process_first(desc="dataset map pre-processing"):
            column_names = dataset.column_names
            splits = dataset.map(_preprocess, batched=True, remove_columns=column_names, desc=desc)
    else:
        column_names = dataset.column_names
        splits = dataset.map(_preprocess, batched=True, remove_columns=column_names, desc=desc)

    return splits


class Preprocessor:
    def __init__(
            self,
            tokenizer: PreTrainedTokenizer,
            ent_start_token: str = "[START_ENT]",
            ent_end_token: str = "[END_ENT]",
            entity_token: str = "[ENT]",
            max_context_length: int = 250,
            max_candidate_length: int = 100,
            context_window_chars: int = 1000,
        ) -> None:
        self.tokenizer = tokenizer
        self.ent_start_token = ent_start_token
        self.ent_end_token = ent_end_token
        self.entity_token = entity_token
        self.max_context_length = max_context_length
        self.max_candidate_length = max_candidate_length
        self.context_window_chars = context_window_chars

        self.start_marker_id = self.tokenizer.convert_tokens_to_ids(self.ent_start_token)
        self.end_marker_id = self.tokenizer.convert_tokens_to_ids(self.ent_end_token)

        # Prepare prefix and suffix ids for special tokens
        if self.tokenizer.cls_token:
            self.prefix_ids = [self.tokenizer.cls_token_id]
        else:
            dummy_id = -1
            with_special_tokens = self.tokenizer.build_inputs_with_special_tokens([dummy_id])
            dummy_idx = with_special_tokens.index(dummy_id)
            self.prefix_ids = with_special_tokens[:dummy_idx]

        if self.tokenizer.sep_token:
            self.suffix_ids = [self.tokenizer.sep_token_id]
        else:
            dummy_id = -1
            with_special_tokens = self.tokenizer.build_inputs_with_special_tokens([dummy_id])
            dummy_idx = with_special_tokens.index(dummy_id)
            self.suffix_ids = with_special_tokens[dummy_idx + 1 :]

        self.offset_correction = len(self.prefix_ids)
        self.max_context_length = self.max_context_length - len(self.prefix_ids) - len(self.suffix_ids)

    def process_context(self, text: str, start: int, end: int, candidate_ids: Optional[list[str]] = None, labels: Optional[list[str]] = None) -> BatchEncoding:
        context, new_start, new_end = cut_context_window(text, start, end, self.context_window_chars)
        marked_text = context[:new_start] + self.ent_start_token + context[new_start:new_end] + self.ent_end_token + context[new_end:]
        input_ids = self.tokenizer.encode(marked_text, add_special_tokens=False, truncation=False)

        start_marker_idx = input_ids.index(self.start_marker_id)
        end_marker_idx = input_ids.index(self.end_marker_id)

        truncated_input_ids, mention_start, mention_end = truncate_around_mention(input_ids, self.max_context_length, start_marker_idx, end_marker_idx + 1)
        final_input_ids = self.prefix_ids + truncated_input_ids + self.suffix_ids
        final_start = mention_start + self.offset_correction
        final_end = mention_end + self.offset_correction

        encodings = BatchEncoding({
            "input_ids": final_input_ids,
            "attention_mask": [1] * len(final_input_ids),
            "start_positions": final_start,
            "end_positions": final_end,
        })
        if labels is not None:
            encodings["labels"] = labels
        if candidate_ids is not None:
            encodings["candidates"] = candidate_ids
        return encodings

    def process_candidate(self, name: str, description: str) -> BatchEncoding:
        marked_text = name + self.entity_token + description
        encodings = self.tokenizer.encode(
            marked_text,
            max_length=self.max_candidate_length,
            add_special_tokens=False
        )
        return encodings

    def data_flatten(self, dataset: Dataset) -> Dataset:
        def _preprocess_example(examples: Dataset) -> Iterator[dict]:
            for id_, text, entities in zip(examples["id"], examples["text"], examples["entities"]):
                for ent in entities:
                    if not ent["label"]:
                        continue
                    yield {"id": id_, "text": text, "entity": ent}
        return preprocess(dataset, _preprocess_example, desc="Filtering")

    def dataset_preprocess(self, dataset: Dataset, candidates_ids: Optional[list[list[str]]] = None) -> Dataset:
        def _preprocess_example(examples: Dataset) -> Iterator[BatchEncoding]:
            for i, (text, entity) in enumerate(zip(examples["text"], examples["entity"])):
                candidate_ids = examples["candidates"][i] if "candidates" in examples else None
                encodings = self.process_context(text, entity['start'], entity['end'], candidate_ids, entity['label'] if 'label' in entity else None)
                yield encodings
        flatten_dataset = self.data_flatten(dataset)
        if candidates_ids is not None:
            assert len(flatten_dataset) == len(candidates_ids)
            flatten_dataset = flatten_dataset.add_column("candidates", candidates_ids)
        processed_dataset = preprocess(flatten_dataset, _preprocess_example, desc="Preprocessing")
        return processed_dataset

    def dictionary_preprocess(self, dictionary: EntityDictionary) -> EntityDictionary:
        def preprocess_example(name: str, description: str) -> BatchEncoding:
            encodings  = self.process_candidate(name, description)
            return encodings
        dictionary.add_encoding(preprocess_example)
        return dictionary



import json
import os
import pickle
import random
import re
from logging import getLogger
from typing import Optional

import numpy as np
from tqdm.auto import tqdm

from entity_linkings.data_utils import EntityDictionary

from ..base import IndexerBase

logger = getLogger(__name__)
logger.setLevel("INFO")

punc_remover = re.compile(r"[\W]+")


def build_simpler_mentions_dict(mention_entities_counter: dict[str, dict[str, int]]) -> dict[str, dict[str, int]]:
    simpler_mentions_candidate_dict: dict[str, dict[str, int]] = {}
    for mention in mention_entities_counter:
        # create mention without blanks and lower cased
        simplified_mention = mention.replace(' ', '').lower()
        # the simplified mention already occurred from another mention
        if simplified_mention in simpler_mentions_candidate_dict:
            for entity in mention_entities_counter[mention]:
                if entity in simpler_mentions_candidate_dict[simplified_mention]:
                    simpler_mentions_candidate_dict[simplified_mention][entity] += mention_entities_counter[mention][entity]
                else:
                    simpler_mentions_candidate_dict[simplified_mention][entity] = mention_entities_counter[mention][entity]
        # its the first occurrence of the simplified mention
        else:
            simpler_mentions_candidate_dict[simplified_mention] = mention_entities_counter[mention]
    return simpler_mentions_candidate_dict


def build_most_simpler_mentions_dict(mention_entities_counter: dict[str, dict[str, int]]) -> dict[str, dict[str, int]]:
    even_more_simpler_mentions_candidate_dict: dict[str, dict[str, int]] = {}
    for mention in mention_entities_counter:
        # create simplified mention
        simplified_mention=punc_remover.sub("", mention.lower())
        # the simplified mention already occurred from another mention
        if simplified_mention in even_more_simpler_mentions_candidate_dict:
            for entity in mention_entities_counter[mention]:
                if entity in even_more_simpler_mentions_candidate_dict[simplified_mention]:
                    even_more_simpler_mentions_candidate_dict[simplified_mention][entity] += mention_entities_counter[mention][entity]
                else:
                    even_more_simpler_mentions_candidate_dict[simplified_mention][entity] = mention_entities_counter[mention][entity]
        # its the first occurrence of the simplified mention
        else:
            even_more_simpler_mentions_candidate_dict[simplified_mention] = mention_entities_counter[mention]
    return even_more_simpler_mentions_candidate_dict


class MentionPriorIndexer(IndexerBase):
    def __init__(self, dictionary: EntityDictionary, mention_counter_path: Optional[str] = None) -> None:
        super().__init__(dictionary)
        self.mention_counter_path = mention_counter_path

    def build_index(self, index_path: str) -> None:
        if os.path.exists(os.path.join(index_path, "mention_entities_counter.pickle")):
            logger.info(f"Loading existing index from {index_path}")
            self.load(index_path=index_path)
        else:
            if self.mention_counter_path is None:
                raise ValueError("mention_counter_path must be provided to build the index.")
            counter = json.load(open(self.mention_counter_path, 'r'))
            self.meta_ids_to_keys = {entity["name"]: entity["id"] for entity in self.dictionary}
            self.mention_id_counter: dict[str, dict[str, int]] = {}
            pbar = tqdm(total=len(counter), desc='Building index')
            for mention, entity_counter in counter.items():
                pbar.update()
                mention_counter = {}
                for entity_name, count in entity_counter.items():
                    entity_name = entity_name.replace(' ', '_')
                    if entity_name in self.meta_ids_to_keys:
                        wiki_id = self.dictionary.get_from_title(entity_name)['id']
                        mention_counter[wiki_id] = count
                self.mention_id_counter[mention] = mention_counter
            pbar.close()
            self.simpler_mentions_candidate_dict = build_simpler_mentions_dict(self.mention_id_counter)
            self.even_more_simpler_mentions_candidate_dict = build_most_simpler_mentions_dict(self.mention_id_counter)

    def search_knn(self, query: str|list[str], top_k: int, ignore_ids: Optional[list[str]|list[list[str]]] = None) -> tuple[np.ndarray, list[list[str]]]:
        if top_k <= 0:
            raise RuntimeError("K is zero or under zero.")
        if top_k > self.num_entities:
            top_k = self.num_entities
            logger.warning(f"K is over size of dictionary. K is modified to size of dictionary to {self.num_entities}")

        queries = [query] if type(query) is str else query
        scores, indices_keys = [], []

        is_per_query_ignore = False
        global_ignore_set = set()
        if ignore_ids is not None:
            if isinstance(ignore_ids[0], list):
                is_per_query_ignore = True
            else:
                global_ignore_set = set(ignore_ids)

        for i, query in enumerate(queries):
            if ignore_ids is None:
                current_ignore_set = set()
            elif is_per_query_ignore:
                current_ignore_set = set(ignore_ids[i])
            else:
                current_ignore_set = global_ignore_set.copy()

            candidates = self.mention_id_counter.get(query, None)
            if not candidates:
                candidates = self.simpler_mentions_candidate_dict.get(query.lower().replace(' ', ''), None)
            if not candidates:
                candidates = self.even_more_simpler_mentions_candidate_dict.get(punc_remover.sub("", query.lower()), None)

            final_ids, final_scores = [], []
            if candidates:
                valid_candidate_ids = [(k, v) for k, v in candidates.items() if k not in current_ignore_set]
                topk_items = sorted(valid_candidate_ids, key=lambda x: x[1], reverse=True)[:top_k]
                if topk_items:
                    num_mentions = sum([x[1] for x in topk_items])
                    final_ids = [x[0] for x in topk_items]
                    final_scores = [x[1] / num_mentions for x in topk_items]
                    current_ignore_set.update(final_ids)

            num_needed = top_k - len(final_ids)
            if num_needed > 0:
                if not candidates:
                    logger.info(f'No candidates found for mention "{query}".')
                final_scores.extend([0.0] * num_needed)
                sample_size = min(num_needed * 2 + 50, self.num_entities)
                while len(final_ids) < top_k:
                    random_samples = random.sample(self.entity_ids, sample_size)
                    for r_s in random_samples:
                        if r_s not in current_ignore_set:
                            final_ids.append(r_s)
                            current_ignore_set.add(r_s)
                            if len(final_ids) == top_k:
                                break
            indices_keys.append(final_ids)
            scores.append(final_scores)

        return np.array(scores), indices_keys

    def load(self, index_path: str) -> None:
        logger.info("Deserializing index from %s", index_path)
        mention_entities_counter_file = os.path.join(index_path, "mention_entities_counter.pickle")
        simpler_mentions_candidate_file = os.path.join(index_path, "simpler_mentions_candidate_dict.pickle")
        even_more_simpler_mentions_candidate_file = os.path.join(index_path, "even_more_simpler_mentions_candidate_dict.pickle")
        meta_file = os.path.join(index_path, "meta_candidate_list.json")
        self.mention_id_counter = pickle.load(open(mention_entities_counter_file, 'rb'))
        self.simpler_mentions_candidate_dict = pickle.load(open(simpler_mentions_candidate_file, 'rb'))
        self.even_more_simpler_mentions_candidate_dict = pickle.load(open(even_more_simpler_mentions_candidate_file, 'rb'))
        self.meta_ids_to_keys = json.load(open(meta_file))
        logger.info(
            "Loaded index of type %s and size %d", type(self.mention_id_counter), len(self.meta_ids_to_keys)
        )

    def save_index(self, index_path: str, ensure_ascii: bool = False) -> None:
        if not os.path.isdir(index_path):
            os.mkdir(index_path)
        logger.info("Serializing index to %s", index_path)
        mention_entities_counter_file = os.path.join(index_path, "mention_entities_counter.pickle")
        simpler_mentions_candidate_file = os.path.join(index_path, "simpler_mentions_candidate_dict.pickle")
        even_more_simpler_mentions_candidate_file = os.path.join(index_path, "even_more_simpler_mentions_candidate_dict.pickle")
        meta_file = os.path.join(index_path, "meta_candidate_list.json")
        pickle.dump(self.mention_id_counter, open(mention_entities_counter_file, 'wb'))
        pickle.dump(self.simpler_mentions_candidate_dict, open(simpler_mentions_candidate_file, 'wb'))
        pickle.dump(self.even_more_simpler_mentions_candidate_dict, open(even_more_simpler_mentions_candidate_file, 'wb'))
        json.dump(self.meta_ids_to_keys, open(meta_file, 'w'), ensure_ascii=ensure_ascii)

    def __len__(self) -> int:
        return len(self.dictionary)



from .base import RetrieverBase
from .bm25 import BM25
from .dualencoder import DUALENCODER
from .e5bm25 import E5BM25
from .prior import PRIOR
from .textembedding import TEXTEMBEDDING

RETRIEVER_CLS: list[type[RetrieverBase]] = [
    BM25,
    PRIOR,
    TEXTEMBEDDING,
    E5BM25,
    DUALENCODER,
]

RETRIEVER_ID2CLS = {
    c.__name__.lower(): c for c in RETRIEVER_CLS
}

__all__ = [
    "RetrieverBase",
    "RETRIEVER_ID2CLS",
]



import logging
import os
from argparse import ArgumentParser, Namespace

from datasets import load_dataset

from entity_linkings import get_retrievers, load_dictionary
from entity_linkings.trainer import TrainingArguments
from entity_linkings.utils import read_yaml

logger = logging.getLogger(__name__)


def main(args: Namespace) -> None:
    if args.gpus is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpus

    dictionary = load_dictionary(args.dictionary_id_or_path, cache_dir=args.cache_dir)
    dataset_id = args.dataset_id if args.dataset_id else "json"
    if dataset_id != "json":
        dataset_id = dataset_id if dataset_id.startswith("naist-nlp/") else f"naist-nlp/{dataset_id}"
        dataset = load_dataset(dataset_id, cache_dir=args.cache_dir)
    else:
        data_files = {"train": args.train_file}
        if args.validation_file:
            data_files["validation"] = args.validation_file
        dataset = load_dataset("json", data_files=data_files, cache_dir=args.cache_dir)
    if args.remove_nil:
        from entity_linkings.data_utils import filter_nil_entities
        dataset["train"] = filter_nil_entities(dataset["train"], dictionary)
        if "validation" in dataset:
            dataset["validation"] = filter_nil_entities(dataset["validation"], dictionary)

    if args.retriever_config is not None:
        all_config = read_yaml(args.retriever_config)
        model_config = all_config[args.retriever_id.lower()]
        training_config = all_config["training_arguments"]
    else:
        model_config = {}
        training_config = {}
    if args.retriever_model_name_or_path is not None:
        model_config["model_name_or_path"] = args.retriever_model_name_or_path

    training_config["output_dir"] = args.output_dir
    training_config["num_train_epochs"] = args.num_train_epochs
    training_config["per_device_train_batch_size"] = args.train_batch_size
    training_config["per_device_eval_batch_size"] = args.validation_batch_size
    training_config["gradient_accumulation_steps"] = args.gradient_accumulation_steps

    training_args = TrainingArguments(**training_config)
    if "validation" not in dataset:
        training_args.eval_strategy = 'no'
        training_args.save_strategy = 'no'

    if args.wandb:
        import wandb
        wandb.init(
            project=os.environ.get("WANDB_PROJECT", "entity_linkings"),
            name=args.retriever_id, tags=["training"]
        )
        wandb.log({
            "model_type": "candidate_retriever",
            "model_name_or_path": args.retriever_model_name_or_path,
            "dictionary_id_or_path": args.dictionary_id_or_path,
            "dataset_id": dataset_id,
            "remove_nil": args.remove_nil,
            "num_hard_negatives": args.num_hard_negatives,
        })
        training_args.report_to = ["wandb"]
        training_args.run_name = args.retriever_id

    retriever_cls = get_retrievers(args.retriever_id)
    retriever = retriever_cls(dictionary, config=retriever_cls.Config(**model_config), index_path=args.retriever_index_dir)
    _ = retriever.train(
        train_dataset=dataset["train"],
        eval_dataset=dataset["validation"] if "validation" in dataset else None,
        training_args=training_args,
        num_hard_negatives=args.num_hard_negatives
    )


def cli_main() -> None:
    parser = ArgumentParser()
    parser.add_argument('--retriever_id', '-m', type=str, default=None, help='Name of the model to use.')
    parser.add_argument('--retriever_model_name_or_path', type=str, default=None, help='Name of the model to use.')
    parser.add_argument('--retriever_index_dir', type=str, default=None, help='Path to the retriever index directory.')
    parser.add_argument('--retriever_config', type=str, default=None, help='YAML-based config file.')
    parser.add_argument('--dictionary_id_or_path', '-d', type=str, default=None, help='Path to the entity dictionary file.')
    parser.add_argument('--dataset_id', '-D', type=str, default=None, help='Name of the dataset to use.')
    parser.add_argument('--train_file', type=str, default=None, help='Path to the training dataset file.')
    parser.add_argument('--validation_file', type=str, default=None, help='Path to the validation dataset file.')
    parser.add_argument('--num_hard_negatives', type=int, default=0, help='Number of hard negatives to use during training.')
    parser.add_argument('--num_train_epochs', type=int, default=5, help='Number of training epochs.')
    parser.add_argument('--train_batch_size', type=int, default=8, help='Training batch size.')
    parser.add_argument('--validation_batch_size', type=int, default=8, help='Validation batch size.')
    parser.add_argument('--gradient_accumulation_steps', type=int, default=1, help='Number of gradient accumulation steps.' )
    parser.add_argument('--output_dir', type=str, required=True, help='Path to the output directory.')
    parser.add_argument('--remove_nil', action='store_true', default=False, help='Whether to remove nil entities from the dataset.')
    parser.add_argument('--cache_dir', type=str, default=None, help='Path to the cache directory.')
    parser.add_argument('--gpus', type=str, default=None, help='Comma-separated list of GPU device IDs to use.')
    parser.add_argument('--wandb', action='store_true', default=False, help='Whether to use wandb for logging.')
    args = parser.parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()
